"""
ExperimentalMatrix
===================
ExperimentalMatrix describes an experiment.

"""

# Python
from __future__ import print_function
import os
import sys
from collections import *

# Internal
from rgt.GenomicRegionSet import *
from rgt.GeneSet import *
# from rgt.helper import pretty

# External

possible_types=["genes","regions","reads"]

class ExperimentalMatrix:
    """ Describes an experimental matrix.

    *Variables:*

        - names -- The unique name of experiment (filename).
        - types -- The type of data.
        - files -- The path of the related file with its filename as keys.
        - fields -- List types of informations including names, types, files and others.
        - fieldsDict -- Its keys are just self.fields, and the values are extra informations.
        - objectsDict -- Key is the names; value is GenomicRegionSet or GeneSet.
        - trash -- List of names being deleted
    """

    def __init__(self):
        self.names = []
        self.types = []
        self.files = {}
        self.fields = []
        self.fieldsDict = {}
        self.objectsDict = {}
        self.trash = []
        
    def read(self, file_path, is_bedgraph=False, verbose=False, test=False):
        """Read Experimental matrix file.

        *Keyword arguments:*

            - file_path -- Experimental matrix file path + name.
            - is_bedgraph -- Whether regions are in bedgraph format (default = False).
            - verbose -- Verbose output (default = False).
            - test -- Fetch only 10 regions form each BED files for test.

        *Example of experimental matrix file:*

            ======== ======== ========= ==================
              name      type     file     further1
            ======== ======== ========= ==================
            MPP_PU1  regions  file1.bed  addidional_info1
            CDP_PU1  regions  file2.bed  addidional_info2
            [ ... ]
            ======== ======== ========= ==================
        """
        f = open(file_path,'rU')

        base_dir = ""
        for line in f:
            # Neglect comment lines
            line = line.strip()
            if not line: continue
            elif line[0] == "#": continue
            
            # Read header
            elif line[:4] == "name":
                header = line.split()
                assert(header[0] == "name")
                assert(header[1] == "type")
                assert(header[2] == "file")
                self.fields = header
                
                #initialize further header files        
                for fi in range(3,len(header)):
                    self.fieldsDict[ header[fi] ] = OrderedDict()
                
            # Read further information    
            else:
                line = line.split()
                if line[0].startswith("BASE_DIR"):
                    base_dir = line[1]
                if len(line) < 3:  # Skip the row which has insufficient information
                    #print("Ignore line, as tab-separated number of fields < 3s: %s" %line, file=sys.stderr)
                    continue
                if verbose: print("Reading: ", line, file=sys.stderr)
                
                self.names.append(line[0])
                self.files[line[0]] = os.path.join(base_dir,line[2]) #dict: filename -> filepath
                self.types.append(line[1])
                
                curr_id = None
                for fi in range(3, len(self.fields)): #read further information
                    d = self.fieldsDict[ self.fields[fi] ]
                    # print(line[fi])
                    if "," in line[fi] and "(" not in line[fi]:
                        for t in line[fi].split(","):
                            try: d[t].append(line[0]+t)
                            except: d[t] = [line[0]+t]
                            self.names.append(line[0]+t)
                            self.files[line[0]+t] = line[2]
                            self.types.append(line[1])
                            for f in range(3, len(self.fields)):
                                if f != fi:
                                    try: self.fieldsDict[ self.fields[f] ][line[f]].append(line[0]+t)
                                    except: self.fieldsDict[ self.fields[f] ][line[f]] = [line[0]+t]
                            if line[0] not in self.trash: 
                                self.trash.append(line[0])
                            
                    else:
                        if curr_id:
                            try: d[line[fi]] += curr_id
                            except: d[line[fi]] = curr_id
                        else:
                            try: d[line[fi]].append(line[0])
                            except: d[line[fi]] = [line[0]]
                            
        # self.types = numpy.array(self.types)
        # self.names = numpy.array(self.names)
        self.remove_name()
        self.load_bed_url(".")
        self.load_objects(is_bedgraph, verbose=verbose, test=test)

        
    def get_genesets(self):
        """Returns the GeneSets."""
        return [self.objectsDict[n] for i,n in enumerate(self.names) if self.types[i] =="genes"]

    def get_regionsets(self):
        """Returns the RegionSets."""
        return [self.objectsDict[n] for i,n in enumerate(self.names) if self.types[i] =="regions"]
    
    def get_regionsnames(self):
        """Returns the region names."""
        return [n for i,n in enumerate(self.names) if self.types[i] =="regions"]
    
    def get_readsfiles(self):
        """Returns the 'read' type files."""
        return [self.files[n] for i,n in enumerate(self.names) if self.types[i]=="reads"]

    def get_readsnames(self):
        """Returns the 'read' type names."""
        return [n for i,n in enumerate(self.names) if self.types[i]=="reads"]

    def load_objects(self, is_bedgraph, verbose=False, test=False):
        """Load files and initialize object.

        *Keyword arguments:*

            - is_bedgraph -- Whether regions are in bedgraph format (default = False).
            - verbose -- Verbose output (default = False).
            - test -- Fetch only 10 regions form each BED files for test.
        """
        for i, t in enumerate(self.types):
            if verbose: print("Loading file ", self.files[self.names[i]], file = sys.stderr)
            
            if t not in ["regions", "genes"] and verbose:
                print("Cannot load objects", file=sys.stderr)
            
            if t == "regions":
                regions = GenomicRegionSet(self.names[i])
                if is_bedgraph:
                    regions.read_bedgraph(os.path.abspath(self.files[self.names[i]]))
                else:
                    regions.read_bed(os.path.abspath(self.files[self.names[i]]))
                    if test: regions.sequences = regions.sequences[0:11]
                self.objectsDict[self.names[i]] = regions
            
            elif t == "genes":
                genes = GeneSet(self.names[i])
                genes.read(os.path.abspath(self.files[self.names[i]]))  # Here change the relative path into absolute path
                self.objectsDict[self.names[i]] = genes
            
    def get_type(self,name,field):
        """Return the type according to the given name and field.

        *Keyword arguments:*

            - name -- Name to return.
            - field -- Field to return.
        """
        
        if field == "regions" or field == "reads":
            field = "factor"
        
        for t in self.fieldsDict[field].keys():
            if name in self.fieldsDict[field][t]:
                return t
        
    def get_types(self,name,skip_all=False):
        """Fetch all extra informations as a list according to the given name.

        *Keyword arguments:*

            - name -- Name to return.
        """
        result = []
        for f in self.fieldsDict.keys():
            for t in self.fieldsDict[f].keys():
                if skip_all and t == "ALL": continue
                elif name in self.fieldsDict[f][t]:
                    result.append(t)
        return result

    def remove_name(self):
        """Removes experiments by name.

        *Keyword arguments:*

            - name -- Name to remove.
        """
        self.trash = list(set(self.trash))
        # for i, name in enumerate(self.names):
        for name in self.trash:
            i = self.names.index(name)
            del self.types[i]
            del self.names[i]
            self.files.pop(name, None)

            for f in self.fieldsDict.keys():
                for t in self.fieldsDict[f].keys(): 
                    # try:
                    if name in self.fieldsDict[f][t]:
                        self.fieldsDict[f][t].remove(name)
                    if not self.fieldsDict[f][t]:
                        self.fieldsDict[f].pop(t, None)
            try: self.objectsDict.pop(name, None)
            except: pass
        self.trash = []

    def match_ms_tags(self,field, test=False):
        """Add more entries to match the missing tags of the given field. For example, there are tags for cell like 'cell_A' and 'cell_B' for reads, but no these tag for regions. Then the regions are repeated for each tags from reads to match all reads.

        *Keyword arguments:*

            - field -- Field to add extra entries.
        """

        # check regions or reads have empty tag
        altypes = self.fieldsDict[field].keys()
        if "ALL" in altypes:
            altypes.remove("ALL")
            for name in self.fieldsDict[field]["ALL"]:
                i = self.names.index(name)
                for t in altypes:
                    # print("\t"+t)
                    n = name+"_"+t
                    # print("\t\t"+n)
                    self.names.append(n)
                    self.types.append(self.types[i])
                    self.files[n] = self.files[name]
                    # types = self.get_types(name,skip_all=True)
                    # print("************")
                    # print(types)

                    for f in self.fields[3:]:
                        if f == field: 
                            try: self.fieldsDict[f][t].append(n)
                            except: self.fieldsDict[f][t] = [n]
                        else:
                            try: self.fieldsDict[f][self.get_type(name=name,field=f)].append(n)
                            except: self.fieldsDict[f][self.get_type(name=name,field=f)] = [n]
                    # for f in self.fieldsDict.keys():
                    #     for ty in types:
                    #         try: self.fieldsDict[f][ty].append(n)
                    #         except: pass
                    if self.types[i] == "regions":
                        g = GenomicRegionSet(n)
                        g.read_bed(self.files[name])
                        if test: g.sequences = g.sequences[0:11]
                        self.objectsDict[n] = g
                    self.trash.append(name)

    def remove_empty_regionset(self):
        """Remove the entry with zero regions."""
        for r in self.get_regionsnames():
            if len(self.objectsDict[r]) == 0:
                self.trash.append(r)
                print("***Warning: "+r+" has zero regions and is ignored.")
        self.remove_name()


    def load_bed_url(self, temp_dir):
        """Load the BED files which contains url as file path to temporary directory."""
        import urllib
        for key, value in self.files.iteritems():
            if self.types[self.names.index(key)] == "regions" and "http" in value:
                tmpfile = os.path.join(temp_dir, value.split('/')[-1])
                dest_name = tmpfile.partition(".gz")[0]
                if not os.path.isfile(dest_name):
                    if not os.path.isfile(tmpfile):
                        urllib.urlretrieve(value, tmpfile)
                    if value.endswith(".gz"):
                        import gzip
                        with gzip.open(tmpfile, 'rb') as infile:
                            with open(dest_name, 'w') as outfile:
                                for line in infile:
                                    outfile.write(line)
                        os.remove(tmpfile)
                    else:
                        dest_name = tmpfile

                self.files[key] = dest_name

    def add_factor_col(self):
        """Add factor column by the entry name"""
        self.fields.append("factor")
        self.fieldsDict["factor"] = {}
        for n in self.names:
            try:
                self.fieldsDict["factor"][n].append(n)
            except:
                self.fieldsDict["factor"][n] = [n]


"""
%prog [options] CONFIG

THOR detects differential peaks in multiple ChIP-seq profiles associated
with two distinct biological conditions.

Copyright (C) 2014-2016  Manuel Allhoff (allhoff@aices.rwth-aachen.de)

This program comes with ABSOLUTELY NO WARRANTY. This is free 
software, and you are welcome to redistribute it under certain 
conditions. Please see LICENSE file for details.
"""

# Python
from __future__ import print_function
import os
import sys
import pysam
import numpy as np
from math import fabs, log, ceil
from operator import add
from os.path import splitext, basename, join, isfile, isdir, exists
from optparse import OptionParser, OptionGroup
from datetime import datetime

# Internal
from rgt.THOR.postprocessing import merge_delete, filter_deadzones
from MultiCoverageSet import MultiCoverageSet
from rgt.GenomicRegionSet import GenomicRegionSet
from rgt.THOR.get_extension_size import get_extension_size
from rgt.THOR.get_fast_gen_pvalue import get_log_pvalue_new
from input_parser import input_parser
from rgt.Util import which, npath
from rgt import __version__

# External
from numpy import linspace
from scipy.optimize import curve_fit
import matplotlib as mpl
#see http://stackoverflow.com/questions/4931376/generating-matplotlib-graphs-without-a-running-x-server
mpl.use('Agg')
import matplotlib.pyplot as plt


FOLDER_REPORT = None


def merge_output(bamfiles, dims, options, no_bw_files, chrom_sizes):
    for i in range(len(bamfiles)):
        rep = i if i < dims[0] else i - dims[0]
        sig = 1 if i < dims[0] else 2

        temp_bed = npath(options.name + '-s%s-rep%s_temp.bed' % (sig, rep))

        files = [options.name + '-' + str(j) + '-s%s-rep%s.bw' %(sig, rep) for j in no_bw_files]
        if len(no_bw_files) > len(bamfiles):
            files = filter(lambda x: isfile(x), files)
            t = ['bigWigMerge'] + files + [temp_bed]
            c = " ".join(t)
            os.system(c)

            os.system("LC_COLLATE=C sort -k1,1 -k2,2n " + temp_bed + ' > ' + temp_bed +'.sort')

            t = ['bedGraphToBigWig', temp_bed + '.sort', chrom_sizes, options.name + '-s%s-rep%s.bw' % (sig, rep)]
            c = " ".join(t)
            os.system(c)

            for f in files:
                os.remove(f)
            os.remove(temp_bed)
            os.remove(temp_bed + ".sort")
        else:
            ftarget = [options.name + '-s%s-rep%s.bw' %(sig, rep) for j in no_bw_files]
            for i in range(len(ftarget)):
                c = ['mv', files[i], ftarget[i]]
                c = " ".join(c)
                os.system(c)


def _func_quad_2p(x, a, c):
    """Return y-value of y=max(|a|*x^2 + x + |c|, 0),
    x may be an array or a single float"""
    res = []
    if type(x) is np.ndarray:
        for el in x:
            res.append(max(el, fabs(a) * el**2 + el + fabs(c)))
            
        return np.asarray(res)
    else:
        return max(x, fabs(a) * x**2 + x + fabs(c))


def _write_emp_func_data(data, name):
    """Write mean and variance data"""
    assert len(data[0]) == len(data[1])
    f = open(FOLDER_REPORT_DATA + name + '.data', 'w')
    for i in range(len(data[0])):
        print(data[0][i], data[1][i], sep='\t', file=f)
    f.close()


def _plot_func(plot_data, outputdir):
    """Plot estimated and empirical function"""

    maxs = [] #max for x (mean), max for y (var)
    for i in range(2): 
        tmp = np.concatenate((plot_data[0][i], plot_data[1][i])) #plot_data [(m, v, p)], 2 elements
        maxs.append(max(tmp[tmp < np.percentile(tmp, 90)]))

    for i in range(2):
        x = linspace(0, max(plot_data[i][0]), int(ceil(max(plot_data[i][0]))))
        y = _func_quad_2p(x, plot_data[i][2][0], plot_data[i][2][1])
        
        for j in range(2):
            #use matplotlib to plot function and datapoints
            #and save datapoints to files
            ext = 'original'
            if j == 1:
                plt.xlim([0, maxs[0]])
                plt.ylim([0, maxs[1]])
                ext = 'norm'
            ax = plt.subplot(111)
            plt.plot(x, y, 'r', label = 'fitted polynomial') #plot polynom
            plt.scatter(plot_data[i][0], plot_data[i][1], label = 'empirical datapoints') #plot datapoints
            ax.legend()
            plt.xlabel('mean')
            plt.ylabel('variance')
            plt.title('Estimated Mean-Variance Function')
            name = "_".join(['mean', 'variance', 'func', 'cond', str(i), ext])
            _write_emp_func_data(plot_data[i], name)
            plt.savefig(FOLDER_REPORT_PICS + name + '.png')
            plt.close()


def _get_data_rep(overall_coverage, name, debug, sample_size):
    """Return list of (mean, var) points for samples 0 and 1"""
    data_rep = []
    for i in range(2):
        cov = np.asarray(overall_coverage[i]) #matrix: (#replicates X #bins)
        h = np.invert((cov==0).all(axis=0)) #assign True to columns != (0,..,0)
        cov = cov[:,h] #remove 0-columns

        r = np.random.randint(cov.shape[1], size=sample_size)
        r.sort()
        cov = cov[:,r]

        m = list(np.squeeze(np.asarray(np.mean(cov*1.0, axis=0))))
        n = list(np.squeeze(np.asarray(np.var(cov*1.0, axis=0))))
        assert len(m) == len(n)

        data_rep.append(zip(m, n))
        data_rep[i].append((0,0))
        data_rep[i] = np.asarray(data_rep[i])
        
    if debug:
        for i in range(2):
            np.save(str(name) + "-emp-data" + str(i) + ".npy", data_rep[i])
    
    for i in range(2):
        data_rep[i] = data_rep[i][data_rep[i][:,0] < np.percentile(data_rep[i][:,0], 99.75)]
        data_rep[i] = data_rep[i][data_rep[i][:,1] < np.percentile(data_rep[i][:,1], 99.75)]
    
    return data_rep


def _fit_mean_var_distr(overall_coverage, name, debug, verbose, outputdir, report, poisson, sample_size=5000):
    """Estimate empirical distribution (quadr.) based on empirical distribution"""
    done = False
    plot_data = [] #means, vars, paras
        
    while not done:
        data_rep = _get_data_rep(overall_coverage, name, debug, sample_size)
        res = []
        for i in range(2):
            try:
                m = np.asarray(map(lambda x: x[0], data_rep[i])) #means list
                v = np.asarray(map(lambda x: x[1], data_rep[i])) #vars list
                
                if len(m) > 0 and len(v) > 0: 
                    try:
                        p, _ = curve_fit(_func_quad_2p, m, v) #fit quad. function to empirical data
                    except:
                        print("Optimal parameters for mu-var-function not found, get new datapoints", file=sys.stderr)
                        break #restart for loop
                else:
                    p = np.array([0, 1])
                
                res.append(p)
                plot_data.append((m, v, p))
                if i == 1:
                    done = True
            except RuntimeError:
                print("Optimal parameters for mu-var-function not found, get new datapoints", file=sys.stderr)
                break #restart for loop
    
    if report:
        _plot_func(plot_data, outputdir)
    
    if poisson:
        print("Use Poisson distribution as emission", file=sys.stderr)
        p[0] = 0
        p[1] = 0
        res = [np.array([0, 0]), np.array([0, 0])]
    
    return lambda x: _func_quad_2p(x, p[0], p[1]), res

    
def dump_posteriors_and_viterbi(name, posteriors, DCS, states):
    print("Computing info...", file=sys.stderr)
    f = open(name + '-posts.bed', 'w')
    g = open(name + '-states-viterbi.bed', 'w')
    
    for i in range(len(DCS.indices_of_interest)):
        cov1, cov2 = _get_covs(DCS, i)
        p1, p2, p3 = posteriors[i][0], posteriors[i][1], posteriors[i][2]
        chrom, start, end = DCS._index2coordinates(DCS.indices_of_interest[i])
        
        print(chrom, start, end, states[i], cov1, cov2, sep='\t', file=g)
        print(chrom, start, end, max(p3, max(p1,p2)), p1, p2, p3, cov1, cov2, sep='\t', file=f)

    f.close()
    g.close()


def _compute_pvalue((x, y, side, distr)):
    a, b = int(np.mean(x)), int(np.mean(y))
    return -get_log_pvalue_new(a, b, side, distr)


def _get_log_ratio(l1, l2):
    l1, l2 = float(np.sum(np.array(l1))), float(np.sum(np.array(l2)))
    try:
        res = l1/l2
    except:
        return sys.maxint
    
    if res > 0:
        try:
            res = log(res)
            if np.isinf(res):
                return sys.maxint
            return res
        except:
            print('error to compute log ratio', l1, l2, file=sys.stderr)
            return sys.maxint
    else:
        return sys.maxint


def _merge_consecutive_bins(tmp_peaks, distr):
    """Merge consecutive peaks and compute p-value. Return list 
    <(chr, s, e, c1, c2, strand)> and <(pvalue)>"""
    peaks = []
    pvalues = []
    i, j, = 0, 0
    
    while i < len(tmp_peaks):
        j+=1
        c, s, e, c1, c2, strand, strand_pos, strand_neg = tmp_peaks[i]
        v1 = c1
        v2 = c2
        
        tmp_pos = [strand_pos]
        tmp_neg = [strand_neg]
        #merge bins
        while i+1 < len(tmp_peaks) and e == tmp_peaks[i+1][1] and strand == tmp_peaks[i+1][5]:
            e = tmp_peaks[i+1][2]
            v1 = map(add, v1, tmp_peaks[i+1][3])
            v2 = map(add, v2, tmp_peaks[i+1][4])
            tmp_pos.append(tmp_peaks[i+1][6])
            tmp_neg.append(tmp_peaks[i+1][7])
            i += 1
        
        side = 'l' if strand == '+' else 'r'
        pvalues.append((v1, v2, side, distr))
        
        ratio = _get_log_ratio(tmp_pos, tmp_neg)
        peaks.append((c, s, e, v1, v2, strand, ratio))
        i += 1
    
    pvalues = map(_compute_pvalue, pvalues)
    assert len(pvalues) == len(peaks)
    
    return pvalues, peaks
    

def _get_covs(DCS, i, as_list=False):
    """For a multivariant Coverageset, return mean coverage cov1 and cov2 at position i"""
    if not as_list:
        cov1 = int(np.mean(DCS.overall_coverage[0][:, DCS.indices_of_interest[i]]))
        cov2 = int(np.mean(DCS.overall_coverage[1][:, DCS.indices_of_interest[i]]))
    else:
        cov1 = DCS.overall_coverage[0][:,DCS.indices_of_interest[i]]
        cov1 = map(lambda x: x[0], np.asarray((cov1)))
        cov2 = DCS.overall_coverage[1][:,DCS.indices_of_interest[i]]
        cov2 = map(lambda x: x[0], np.asarray((cov2)))
    
    return cov1, cov2


def get_peaks(name, DCS, states, exts, merge, distr, pcutoff, debug, no_correction, deadzones, p=70):
    """Merge Peaks, compute p-value and give out *.bed and *.narrowPeak"""
    exts = np.mean(exts)
    tmp_peaks = []
    tmp_data = []
    
    for i in range(len(DCS.indices_of_interest)):
        if states[i] not in [1,2]:
            continue #ignore background states

        strand = '+' if states[i] == 1 else '-'
        cov1, cov2 = _get_covs(DCS, i, as_list=True)
        
        cov1_strand = np.sum(DCS.overall_coverage_strand[0][0][:,DCS.indices_of_interest[i]]) + np.sum(DCS.overall_coverage_strand[1][0][:,DCS.indices_of_interest[i]])
        cov2_strand = np.sum(DCS.overall_coverage_strand[0][1][:,DCS.indices_of_interest[i]] + DCS.overall_coverage_strand[1][1][:,DCS.indices_of_interest[i]])
        
        chrom, start, end = DCS._index2coordinates(DCS.indices_of_interest[i])
        
        tmp_peaks.append((chrom, start, end, cov1, cov2, strand, cov1_strand, cov2_strand))
        side = 'l' if strand == '+' else 'r'
        tmp_data.append((sum(cov1), sum(cov2), side, distr))
    
    if not tmp_data:
        print('no data', file=sys.stderr)
        return [], [], []
    
    tmp_pvalues = map(_compute_pvalue, tmp_data)
    per = np.percentile(tmp_pvalues, p)
    
    tmp = []
    res = tmp_pvalues > per
    for j in range(len(res)):
        if res[j]:
            tmp.append(tmp_peaks[j])
    tmp_peaks = tmp

    pvalues, peaks, = _merge_consecutive_bins(tmp_peaks, distr) #merge consecutive peaks and compute p-value
    regions = merge_delete(exts, merge, peaks, pvalues) #postprocessing, returns GenomicRegionSet with merged regions
    if deadzones:
        regions = filter_deadzones(deadzones, regions)
    output = []
    pvalues = []
    ratios = []
    main_sep = ':' #sep <counts> main_sep <counts> main_sep <pvalue>
    int_sep = ';' #sep counts in <counts>
    
    for i, el in enumerate(regions):
        tmp = el.data.split(',')
        counts = ",".join(tmp[0:len(tmp)-1]).replace('], [', int_sep).replace('], ', int_sep).replace('([', '').replace(')', '').replace(', ', main_sep)
        pvalue = float(tmp[len(tmp)-2].replace(")", "").strip())
        ratio = float(tmp[len(tmp)-1].replace(")", "").strip())
        pvalues.append(pvalue)
        ratios.append(ratio)
        output.append((el.chrom, el.initial, el.final, el.orientation, counts))
    
    return ratios, pvalues, output


def _output_ext_data(ext_data_list, bamfiles):
    """Output textfile and png file of read size estimation"""
    names = [splitext(basename(bamfile))[0] for bamfile in bamfiles]

    for k, ext_data in enumerate(ext_data_list):
        f = open(FOLDER_REPORT_DATA + 'fragment_size_estimate_' + names[k] + '.data', 'w')
        for d in ext_data:
            print(d[0], d[1], sep='\t', file=f)
        f.close()
    
    for i, ext_data in enumerate(ext_data_list):
        d1 = map(lambda x: x[0], ext_data)
        d2 = map(lambda x: x[1], ext_data)
        ax = plt.subplot(111)
        plt.xlabel('shift')
        plt.ylabel('convolution')
        plt.title('Fragment Size Estimation')
        plt.plot(d2, d1, label=names[i])
    
    ax.legend()
    plt.savefig(FOLDER_REPORT_PICS + 'fragment_size_estimate.png')
    plt.close()


def _compute_extension_sizes(bamfiles, exts, inputs, exts_inputs, report):
    """Compute Extension sizes for bamfiles and input files"""
    start = 0
    end = 600
    ext_stepsize = 5
    
    ext_data_list = []
    #compute extension size
    if not exts:
        print("Computing read extension sizes for ChIP-seq profiles", file=sys.stderr)
        for bamfile in bamfiles:
            e, ext_data = get_extension_size(bamfile, start=start, end=end, stepsize=ext_stepsize)
            exts.append(e)
            ext_data_list.append(ext_data)
    
    if report and ext_data_list:
        _output_ext_data(ext_data_list, bamfiles)
    
    if inputs and not exts_inputs:
        exts_inputs = [5] * len(inputs)

    return exts, exts_inputs


def get_all_chrom(bamfiles):
    chrom = set()
    for bamfile in bamfiles:
        bam = pysam.Samfile(bamfile, "rb" )
        for read in bam.fetch():
            c = bam.getrname(read.reference_id)
            if c not in chrom:
                chrom.add(c)
    return chrom


def initialize(name, dims, genome_path, regions, stepsize, binsize, bamfiles, exts, \
               inputs, exts_inputs, factors_inputs, chrom_sizes, verbose, no_gc_content, \
               tracker, debug, norm_regions, scaling_factors_ip, save_wig, housekeeping_genes, \
               test, report, chrom_sizes_dict, counter, end, gc_content_cov=None, avg_gc_content=None, \
               gc_hist=None, output_bw=True, save_input=False, m_threshold=80, a_threshold=95, rmdup=False):
    """Initialize the MultiCoverageSet"""
    regionset = regions
    regionset.sequences.sort()
    
    if norm_regions:
        norm_regionset = GenomicRegionSet('norm_regions')
        norm_regionset.read_bed(norm_regions)
    else:
        norm_regionset = None
        
    exts, exts_inputs = _compute_extension_sizes(bamfiles, exts, inputs, exts_inputs, report)
    
    multi_cov_set = MultiCoverageSet(name=name, regions=regionset, dims=dims, genome_path=genome_path,
                                     binsize=binsize, stepsize=stepsize, rmdup=rmdup, path_bamfiles=bamfiles,
                                     path_inputs=inputs, exts=exts, exts_inputs=exts_inputs,
                                     factors_inputs=factors_inputs, chrom_sizes=chrom_sizes, verbose=verbose,
                                     no_gc_content=no_gc_content, chrom_sizes_dict=chrom_sizes_dict, debug=debug,
                                     norm_regionset=norm_regionset, scaling_factors_ip=scaling_factors_ip,
                                     save_wig=save_wig, strand_cov=True, housekeeping_genes=housekeeping_genes,
                                     tracker=tracker, gc_content_cov=gc_content_cov, avg_gc_content=avg_gc_content,
                                     gc_hist=gc_hist, end=end, counter=counter, output_bw=output_bw,
                                     folder_report=FOLDER_REPORT, report=report, save_input=save_input,
                                     m_threshold=m_threshold, a_threshold=a_threshold)
    return multi_cov_set


class HelpfulOptionParser(OptionParser):
    """An OptionParser that prints full help on errors."""
    def error(self, msg):
        self.print_help(sys.stderr)
        self.exit(2, "\n%s: error: %s\n" % (self.get_prog_name(), msg))


def _callback_list(option, opt, value, parser):
    setattr(parser.values, option.dest, map(lambda x: int(x), value.split(',')))


def _callback_list_float(option, opt, value, parser):
    setattr(parser.values, option.dest, map(lambda x: float(x), value.split(',')))


def handle_input():
    parser = HelpfulOptionParser(usage=__doc__)

    parser.add_option("-n", "--name", default=None, dest="name", type="string",
                      help="Experiment's name and prefix for all files that are created.")
    parser.add_option("-m", "--merge", default=False, dest="merge", action="store_true",
                      help="Merge peaks which have a distance less than the estimated mean fragment size "
                           "(recommended for histone data). [default: do not merge]")
    parser.add_option("--housekeeping-genes", default=None, dest="housekeeping_genes", type="str",
                      help="Define housekeeping genes (BED format) used for normalizing. [default: %default]")
    parser.add_option("--output-dir", dest="outputdir", default=None, type="string",
                      help="Store files in output directory. [default: %default]")
    parser.add_option("--report", dest="report", default=False, action="store_true",
                      help="Generate HTML report about experiment. [default: %default]")
    parser.add_option("--deadzones", dest="deadzones", default=None,
                      help="Define blacklisted genomic regions avoided for analysis (BED format). [default: %default]")
    parser.add_option("--no-correction", default=False, dest="no_correction", action="store_true",
                      help="Do not use multipe test correction for p-values (Benjamini/Hochberg). [default: %default]")
    parser.add_option("-p", "--pvalue", dest="pcutoff", default=0.1, type="float",
                      help="P-value cutoff for peak detection. Call only peaks with p-value lower than cutoff. "
                           "[default: %default]")
    parser.add_option("--exts", default=None, dest="exts", type="str", action='callback', callback=_callback_list,
                      help="Read's extension size for BAM files (comma separated list for each BAM file in config "
                           "file). If option is not chosen, estimate extension sizes. [default: %default]")
    parser.add_option("--factors-inputs", default=None, dest="factors_inputs", type="str", action="callback",
                      callback=_callback_list_float,
                      help="Normalization factors for input-DNA (comma separated list for each BAM file in config "
                           "file). If option is not chosen, estimate factors. [default: %default]")
    parser.add_option("--scaling-factors", default=None, dest="scaling_factors_ip", type="str", action='callback',
                      callback=_callback_list_float,
                      help="Scaling factor for each BAM file (not control input-DNA) as comma separated list for "
                           "each BAM file in config file. If option is not chosen, follow normalization strategy "
                           "(TMM or HK approach) [default: %default]")
    parser.add_option("--save-input", dest="save_input", default=False, action="store_true",
                      help="Save input-DNA file if available. [default: %default]")
    parser.add_option("--version", dest="version", default=False, action="store_true",
                      help="Show script's version.")

    group = OptionGroup(parser, "Advanced options")
    group.add_option("--regions", dest="regions", default=None, type="string",
                     help="Define regions (BED format) to restrict the analysis, that is, where to train the HMM and "
                          "search for DPs. It is faster, but less precise.")
    group.add_option("-b", "--binsize", dest="binsize", default=100, type="int",
                     help="Size of underlying bins for creating the signal. [default: %default]")
    group.add_option("-s", "--step", dest="stepsize", default=50, type="int",
                     help="Stepsize with which the window consecutively slides across the genome to create the "
                          "signal. [default: %default]")
    group.add_option("--debug", default=False, dest="debug", action="store_true",
                     help="Output debug information. Warning: space consuming! [default: %default]")
    group.add_option("--no-gc-content", dest="no_gc_content", default=False, action="store_true",
                     help="Do not normalize towards GC content. [default: %default]")
    group.add_option("--norm-regions", default=None, dest="norm_regions", type="str",
                     help="Restrict normalization to particular regions (BED format). [default: %default]")
    group.add_option("-f", "--foldchange", dest="foldchange", default=1.6, type="float",
                     help="Fold change parameter to define training set (t_1, see paper). [default: %default]")
    group.add_option("-t", "--threshold", dest="threshold", default=95, type="float",
                     help="Minimum signal support for differential peaks to define training set as percentage "
                          "(t_2, see paper). [default: %default]")
    group.add_option("--size", dest="size_ts", default=10000, type="int",
                     help="Number of bins the HMM's training set constists of. [default: %default]")
    group.add_option("--par", dest="par", default=1, type="int",
                     help="Percentile for p-value postprocessing filter. [default: %default]")
    group.add_option("--poisson", default=False, dest="poisson", action="store_true",
                     help="Use binomial distribution as emmission. [default: %default]")
    group.add_option("--single-strand", default=False, dest="singlestrand", action="store_true",
                     help="Allow single strand BAM file as input. [default: %default]")
    group.add_option("--m_threshold", default=80, dest="m_threshold", type="int",
                     help="Define the M threshold of percentile for training TMM. [default: %default]")
    group.add_option("--a_threshold", default=95, dest="a_threshold", type="int",
                     help="Define the A threshold of percentile for training TMM. [default: %default]")
    group.add_option("--rmdup", default=False, dest="rmdup", action="store_true",
                     help="Remove the duplicate reads [default: %default]")
    parser.add_option_group(group)

    (options, args) = parser.parse_args()
    options.save_wig = False
    options.exts_inputs = None
    options.verbose = False
    options.hmm_free_para = False

    if options.version:
        print("")
        print(__version__)
        sys.exit()

    if len(args) != 1:
        parser.error("Please give config file")

    config_path = npath(args[0])

    if not isfile(config_path):
        parser.error("Config file %s does not exist!" % config_path)

    bamfiles, genome, chrom_sizes, inputs, dims = input_parser(config_path)

    if not genome:
        options.no_gc_content = True

    if options.exts and len(options.exts) != len(bamfiles):
        parser.error("Number of Extension Sizes must equal number of bamfiles")

    if options.exts_inputs and len(options.exts_inputs) != len(inputs):
        parser.error("Number of Input Extension Sizes must equal number of input bamfiles")

    if options.scaling_factors_ip and len(options.scaling_factors_ip) != len(bamfiles):
        parser.error("Number of scaling factors for IP must equal number of bamfiles")

    for bamfile in bamfiles:
        if not isfile(bamfile):
            parser.error("BAM file %s does not exist!" % bamfile)

    if not inputs and options.factors_inputs:
        print("As no input-DNA, do not use input-DNA factors", file=sys.stderr)
        options.factors_inputs = None

    if options.factors_inputs and len(options.factors_inputs) != len(bamfiles):
        parser.error("factors for input-DNA must equal number of BAM files!")

    if inputs:
        for bamfile in inputs:
            if not isfile(bamfile):
                parser.error("BAM file %s does not exist!" % bamfile)

    if options.regions:
        if not isfile(options.regions):
            parser.error("Region file %s does not exist!" % options.regions)

    if genome and not isfile(genome):
        parser.error("Genome file %s does not exist!" % genome)

    if options.name is None:
        d = str(datetime.now()).replace("-", "_").replace(":", "_").replace(" ", "_").replace(".", "_").split("_")
        options.name = "THOR-exp" + "-" + "_".join(d[:len(d) - 1])

    if not which("wigToBigWig") or not which("bedGraphToBigWig") or not which("bigWigMerge"):
        print("Warning: wigToBigWig, bigWigMerge or bedGraphToBigWig not found! Signal will not be stored!",
              file=sys.stderr)

    if options.outputdir:
        options.outputdir = npath(options.outputdir)
        if isdir(options.outputdir) and sum(
                map(lambda x: x.startswith(options.name), os.listdir(options.outputdir))) > 0:
            parser.error("Output directory exists and contains files with names starting with your chosen experiment "
                         "name! Do nothing to prevent file overwriting!")
        if not exists(options.outputdir):
            os.mkdir(options.outputdir)
    else:
        options.outputdir = os.getcwd()

    options.name = join(options.outputdir, options.name)

    if isdir(join(options.outputdir, 'report_'+basename(options.name))):
        parser.error("Folder 'report_"+basename(options.name)+"' already exits in output directory!" 
                     "Do nothing to prevent file overwriting! "
                     "Please rename report folder or change working directory of THOR with the option --output-dir")

    if options.report:
        os.mkdir(join(options.outputdir, 'report_'+basename(options.name)+"/"))
        os.mkdir(join(options.outputdir, 'report_'+basename(options.name), 'pics/'))
        os.mkdir(join(options.outputdir, 'report_'+basename(options.name), 'pics/data/'))

    global FOLDER_REPORT
    global FOLDER_REPORT_PICS
    global FOLDER_REPORT_DATA
    global OUTPUTDIR
    global NAME

    FOLDER_REPORT = join(options.outputdir, 'report_'+basename(options.name)+"/")
    FOLDER_REPORT_PICS = join(options.outputdir, 'report_'+basename(options.name), 'pics/')
    FOLDER_REPORT_DATA = join(options.outputdir, 'report_'+basename(options.name), 'pics/data/')
    OUTPUTDIR = options.outputdir
    NAME = options.name

    if not inputs:
        print("Warning: Do not compute GC-content, as there is no input file", file=sys.stderr)

    if not genome:
        print("Warning: Do not compute GC-content, as there is no genome file", file=sys.stderr)

    if options.exts is None:
        options.exts = []

    if options.exts_inputs is None:
        options.exts_inputs = []

    return options, bamfiles, genome, chrom_sizes, dims, inputs



###################################################################################################
# Libraries
###################################################################################################

# Python
import os
import warnings

warnings.filterwarnings("ignore")
from itertools import product
from math import floor

# Internal
from rgt.Util import ErrorHandler
from rgt.Util import AuxiliaryFunctions

# External
from pysam import __version__ as ps_version
from pysam import Samfile
from pysam import Fastafile

from Bio import motifs
from Bio.Seq import Seq


###################################################################################################
# Classes
###################################################################################################

class BiasTable:
    """
    Represent a bias table.

    Authors: Eduardo G. Gusmao.
    """

    def __init__(self, output_loc=None):
        """ 
        Initializes BiasTable.
        """
        self.output_loc = output_loc

    def load_table(self, table_file_name_F, table_file_name_R):
        """ 
        Creates a bias table from a tab separated file with a k-mer and bias estimate in each line.

        Keyword arguments:
        table_file_name -- Table file name.
        
        Return:
        bias_table_F, bias_table_R -- Bias tables.
        """
        bias_table_F = dict()
        table_file_F = open(table_file_name_F, "r")
        for line in table_file_F:
            ll = line.strip().split("\t")
            bias_table_F[ll[0]] = float(ll[1])
        table_file_F.close()
        bias_table_R = dict()
        table_file_R = open(table_file_name_R, "r")
        for line in table_file_R:
            ll = line.strip().split("\t")
            bias_table_R[ll[0]] = float(ll[1])
        table_file_R.close()
        return [bias_table_F, bias_table_R]

    def write_tables(self, file_name, table):
        f = open(file_name + "_F.txt", "w")
        for t in table[0].keys():
            f.write(t + "\t" + str(table[0][t]) + "\n")
        f.close()
        f = open(file_name + "_R.txt", "w")
        for t in table[1].keys():
            f.write(t + "\t" + str(table[1][t]) + "\n")
        f.close()

    def estimate_table(self, regions, dnase_file_name, genome_file_name, k_nb, forward_shift, reverse_shift):
        """ 
        Estimates bias based on HS regions, DNase-seq signal and genomic sequences.

        Keyword arguments:
        regions -- DNase-seq HS regions.
        dnase_file_name -- DNase-seq file name.
        genome_file_name -- Genome to fetch genomic sequences from.
        
        Return:
        bias_table_F, bias_table_R -- Bias tables.
        """

        # Parameters
        maxDuplicates = 100
        pseudocount = 1.0

        # Initializing bam and fasta
        if (dnase_file_name.split(".")[-1].upper() != "BAM"): return None  # TODO ERROR
        bamFile = Samfile(dnase_file_name, "rb")
        fastaFile = Fastafile(genome_file_name)

        # Initializing dictionaries
        obsDictF = dict()
        obsDictR = dict()
        expDictF = dict()
        expDictR = dict()

        ct_reads_r = 0
        ct_reads_f = 0
        ct_kmers = 0

        # Iterating on HS regions
        for region in regions:

            # Initialization
            prevPos = -1
            trueCounter = 0

            # Evaluating observed frequencies ####################################
            # Fetching reads
            for r in bamFile.fetch(region.chrom, region.initial, region.final):

                # Calculating positions
                if (not r.is_reverse):
                    cut_site = r.pos + forward_shift - 1
                    p1 = cut_site - int(floor(k_nb / 2))
                else:
                    cut_site = r.aend + reverse_shift + 1
                    p1 = cut_site - int(floor(k_nb / 2))
                p2 = p1 + k_nb

                # Verifying PCR artifacts
                if (p1 == prevPos):
                    trueCounter += 1
                else:
                    prevPos = p1
                    trueCounter = 0
                if (trueCounter > maxDuplicates): continue

                # Fetching k-mer
                try:
                    currStr = str(fastaFile.fetch(region.chrom, p1, p2)).upper()
                except Exception:
                    continue
                if (r.is_reverse): currStr = AuxiliaryFunctions.revcomp(currStr)

                # Counting k-mer in dictionary
                if (not r.is_reverse):
                    ct_reads_f += 1
                    try:
                        obsDictF[currStr] += 1
                    except Exception:
                        obsDictF[currStr] = 1
                else:
                    ct_reads_r += 1
                    try:
                        obsDictR[currStr] += 1
                    except Exception:
                        obsDictR[currStr] = 1

            # Evaluating expected frequencies ####################################
            # Fetching whole sequence
            try:
                currStr = str(fastaFile.fetch(region.chrom, region.initial, region.final)).upper()
            except Exception:
                continue
            currRevComp = AuxiliaryFunctions.revcomp(currStr)

            # Iterating on each sequence position
            for i in range(0, len(currStr) - k_nb):
                ct_kmers += 1
                # Counting k-mer in dictionary
                s = currStr[i:i + k_nb]
                try:
                    expDictF[s] += 1
                except Exception:
                    expDictF[s] = 1

                # Counting k-mer in dictionary for reverse complement
                s = currRevComp[i:i + k_nb]
                try:
                    expDictR[s] += 1
                except Exception:
                    expDictR[s] = 1

        # Closing files
        bamFile.close()
        fastaFile.close()

        # Creating bias dictionary
        alphabet = ["A", "C", "G", "T"]
        kmerComb = ["".join(e) for e in product(alphabet, repeat=k_nb)]
        bias_table_F = dict([(e, 0.0) for e in kmerComb])
        bias_table_R = dict([(e, 0.0) for e in kmerComb])
        for kmer in kmerComb:
            try:
                obsF = obsDictF[kmer] + pseudocount
            except Exception:
                obsF = pseudocount
            try:
                expF = expDictF[kmer] + pseudocount
            except Exception:
                expF = pseudocount
            if ct_reads_f == 0:
                bias_table_F[kmer] = 1
            else:
                bias_table_F[kmer] = round(float(obsF / ct_reads_f) / float(expF / ct_kmers), 6)
            try:
                obsR = obsDictR[kmer] + pseudocount
            except Exception:
                obsR = pseudocount
            try:
                expR = expDictR[kmer] + pseudocount
            except Exception:
                expR = pseudocount
            if ct_reads_r == 0:
                bias_table_R[kmer] = 1
            else:
                bias_table_R[kmer] = round(float(obsR / ct_reads_r) / float(expR / ct_kmers), 6)

        # Return
        return [bias_table_F, bias_table_R]

    def get_pwm_score(self, sequence, pwm, k_nb):
        score = 1.0
        for position in range(k_nb):
            letter = sequence[position]
            score *= pwm[letter][position]
        return score

    def estimate_table_pwm(self, regions, dnase_file_name, genome_file_name, k_nb, forward_shift, reverse_shift):
        """
        Estimates bias based on HS regions, DNase-seq signal and genomic sequences.

        Keyword arguments:
        regions -- DNase-seq HS regions.
        atac_file_name -- DNase-seq file name.
        genome_file_name -- Genome to fetch genomic sequences from.

        Return:
        bias_table_F, bias_table_R -- Bias tables.
        """

        # Initializing bam and fasta
        if (dnase_file_name.split(".")[-1].upper() != "BAM"): return None  # TODO ERROR
        bamFile = Samfile(dnase_file_name, "rb")
        fastaFile = Fastafile(genome_file_name)

        obsSeqsF = []
        obsSeqsR = []
        expSeqsF = []
        expSeqsR = []

        # Iterating on HS regions
        for region in regions:
            # Evaluating observed frequencies
            # Fetching reads
            for r in bamFile.fetch(region.chrom, region.initial, region.final):
                # Calculating positions
                # if(not r.is_reverse): p1 = r.pos - (k_nb/2) - 1 + shift
                # else: p1 = r.aend - (k_nb/2) + 1 - shift
                if (not r.is_reverse):
                    cut_site = r.pos + forward_shift - 1
                    p1 = cut_site - int(floor(k_nb / 2))
                else:
                    cut_site = r.aend + reverse_shift + 1
                    p1 = cut_site - int(floor(k_nb / 2))
                p2 = p1 + k_nb

                # Fetching k-mer
                try:
                    currStr = str(fastaFile.fetch(region.chrom, p1, p2)).upper()
                except Exception:
                    continue
                if (r.is_reverse): currStr = AuxiliaryFunctions.revcomp(currStr)

                # Counting k-mer in dictionary
                if 'N' not in currStr:
                    if (not r.is_reverse):
                        obsSeqsF.append(Seq(currStr))
                    else:
                        obsSeqsR.append(Seq(currStr))

            # Evaluating expected frequencies
            # Fetching whole sequence
            try:
                currStr = str(fastaFile.fetch(region.chrom, region.initial, region.final)).upper()
            except Exception:
                continue
            currRevComp = AuxiliaryFunctions.revcomp(currStr)

            # Iterating on each sequence position
            for i in range(0, len(currStr) - k_nb):
                s = currStr[i:i + k_nb]
                if 'N' not in currStr:
                    # Counting k-mer in dictionary
                    expSeqsF.append(Seq(s))

                    # Counting k-mer in dictionary for reverse complement
                    s = currRevComp[i:i + k_nb]
                    expSeqsR.append(Seq(s))

        # Closing files
        bamFile.close()
        fastaFile.close()

        obsMotifsF = motifs.create(obsSeqsF)
        obsMotifsR = motifs.create(obsSeqsR)
        expMotifsF = motifs.create(expSeqsF)
        expMotifsR = motifs.create(expSeqsR)

        obsPwmF = obsMotifsF.pwm
        obsPwmR = obsMotifsR.pwm
        expPwmF = expMotifsF.pwm
        expPwmR = expMotifsR.pwm

        # Output logos
        logo_obs_f = os.path.join(self.output_loc, "Bias", "logo",
                                       "obs_{}_{}_f.pdf".format(str(k_nb), str(forward_shift)))
        logo_obs_r = os.path.join(self.output_loc, "Bias", "logo",
                                       "obs_{}_{}_r.pdf".format(str(k_nb), str(forward_shift)))
        logo_exp_f = os.path.join(self.output_loc, "Bias", "logo",
                                       "exp_{}_{}_f.pdf".format(str(k_nb), str(forward_shift)))
        logo_exp_r = os.path.join(self.output_loc, "Bias", "logo",
                                       "exp_{}_{}_r.pdf".format(str(k_nb), str(forward_shift)))
        obsMotifsF.weblogo(logo_obs_f, format="pdf", stack_width="large", color_scheme="color_classic",
                           yaxis_scale=0.2, yaxis_tic_interval=0.1)
        obsMotifsR.weblogo(logo_obs_r, format="pdf", stack_width="large", color_scheme="color_classic",
                           yaxis_scale=0.2, yaxis_tic_interval=0.1)
        expMotifsF.weblogo(logo_exp_f, format="pdf", stack_width="large", color_scheme="color_classic",
                           yaxis_scale=0.02, yaxis_tic_interval=0.01)
        expMotifsR.weblogo(logo_exp_r, format="pdf", stack_width="large", color_scheme="color_classic",
                           yaxis_scale=0.02, yaxis_tic_interval=0.01)

        # Output pwms
        pwm_data_list = [obsPwmF, obsPwmR, expPwmF, expPwmR]
        pwm_file_list = []
        pwm_obs_f = os.path.join(self.output_loc, "Bias", "pwm",
                                       "obs_{}_{}_f.pwm".format(str(k_nb), str(forward_shift)))
        pwm_obs_r = os.path.join(self.output_loc, "Bias", "pwm",
                                       "obs_{}_{}_r.pwm".format(str(k_nb), str(forward_shift)))
        pwm_exp_f = os.path.join(self.output_loc, "Bias", "pwm",
                                       "exp_{}_{}_f.pwm".format(str(k_nb), str(forward_shift)))
        pwm_exp_r = os.path.join(self.output_loc, "Bias", "pwm",
                                       "exp_{}_{}_r.pwm".format(str(k_nb), str(forward_shift)))

        pwm_file_list.append(pwm_obs_f)
        pwm_file_list.append(pwm_obs_r)
        pwm_file_list.append(pwm_exp_f)
        pwm_file_list.append(pwm_exp_r)

        for i in range(len(pwm_data_list)):
            with open(pwm_file_list[i], "w") as f:
                f.write(str(pwm_data_list[i]))

        # Creating bias dictionary
        alphabet = ["A", "C", "G", "T"]
        k_mer_comb = ["".join(e) for e in product(alphabet, repeat=k_nb)]
        bias_table_F = dict([(e, 0.0) for e in k_mer_comb])
        bias_table_R = dict([(e, 0.0) for e in k_mer_comb])
        for k_mer in k_mer_comb:
            obsF = self.get_pwm_score(k_mer, obsPwmF, k_nb)
            expF = self.get_pwm_score(k_mer, expPwmF, k_nb)
            bias_table_F[k_mer] = round(obsF / expF, 6)
            obsR = self.get_pwm_score(k_mer, obsPwmR, k_nb)
            expR = self.get_pwm_score(k_mer, expPwmR, k_nb)
            bias_table_R[k_mer] = round(obsR / expR, 6)

        # Return
        return [bias_table_F, bias_table_R]

# if __name__ == "__main__":
#  import sys
#  from rgt.GenomicRegionSet import *
#  bam_file=sys.argv[1]
#  fasta_file=sys.argv[2]
#  bed_file=sys.argv[3]
#  kmer=int(sys.argv[4])
#  shift=int(sys.argv[5])
#  out=sys.argv[6]
#  regions=GenomicRegionSet("regions")
#  regions.read_bed(bed_file)
#  table=BiasTable(regions=regions,dnase_file_name=bam_file,genome_file_name=fasta_file,k_nb=kmer,shift=shift)
#  table.write_tables(out)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
THOR detects differential peaks in multiple ChIP-seq profiles associated
with two distinct biological conditions.

Copyright (C) 2014-2016 Manuel Allhoff (allhoff@aices.rwth-aachen.de)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

@author: Manuel Allhoff
"""

from __future__ import print_function
import string
# from scipy.stats import binom
from hmmlearn.hmm import _BaseHMM
import sys
# from time import time
from math import fabs, log
from sklearn.utils.extmath import logsumexp

# from scipy.stats import nbinom
# from random import randint
# from scipy.special import gamma
# from operator import mul
#
# import scipy.special as special
# import scipy.optimize as optimize
import numpy as np
from neg_bin import NegBin

import warnings
warnings.filterwarnings('error')

def _get_pvalue_distr(mu, alpha, tracker):
    """Derive NB1 parameters for p-value calculation"""
    mu = mu[0,0]
    alpha = alpha[0,0] / 10000.
    tracker.write(text=str(mu), header="Neg. Bin. distribution for p-value estimates (mu)")
    tracker.write(text=str(alpha), header="Neg. Bin. distribution for p-value estimates (alpha)")
    
    nb = NegBin(mu, alpha)
    return {'distr_name': 'nb', 'distr': nb}

def get_init_parameters(s0, s1, s2, **info):
    """For given training set (s0: Background, s1: Gaining, s2: loseing) get inital mu, alpha for NB1."""
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)
        mu = np.matrix([np.mean(map(lambda x: x[i], s)) for i in range(2) for s in [s0, s1, s2]]).reshape(2, 3)
        var = np.matrix([np.var(map(lambda x: x[i], s)) for i in range(2) for s in [s0, s1, s2]]).reshape(2, 3)
    
    alpha = (var - mu) / np.square(mu)
    alpha[alpha < 0] = 0.001
    
    for el in [mu, alpha]:
        high = min(el[0,1], el[1,2]) + 0.5 * fabs(el[0,1] - el[1,2])
        low = min(el[1,1], el[0,2]) + 0.5 * fabs(el[1,1] - el[0,2])
        med = np.mean([el[0,0], el[1,0]])
        el[0,1] = high
        el[1,2] = high
        el[1,1] = low
        el[0,2] = low
        el[0,0] = med
        el[1,0] = med

    return alpha, mu
    
class NegBinRepHMM(_BaseHMM):
    def __init__(self, alpha, mu, dim_cond_1, dim_cond_2, init_state_seq=None, n_components=3, covariance_type='diag', startprob=[1, 0, 0],
                 transmat=None, startprob_prior=None, transmat_prior=None, func=None,
                 algorithm="viterbi", means_prior=None, means_weight=0,
                 covars_prior=1e-2, covars_weight=1,
                 random_state=None, n_iter=30, thresh=1e-2,
                 params=string.ascii_letters,
                 init_params=string.ascii_letters):
    
        _BaseHMM.__init__(self, n_components, startprob, transmat,
                          startprob_prior=startprob_prior,
                          transmat_prior=transmat_prior, algorithm=algorithm,
                          random_state=random_state, n_iter=n_iter,
                          thresh=thresh, params=params,
                          init_params=init_params)
        
        self.dim = [dim_cond_1, dim_cond_2] #dimension of one emission
        self.n_features = 2 #sum(self.dim) #emission dimension
        self.alpha = alpha
        self.mu = mu
        self._update_distr(self.mu, self.alpha)
        self.func = func
        self.em_prob = 0
    
    
    def fit(self, obs, three_para):
        """Estimate model parameters.

        An initialization step is performed before entering the EM
        algorithm. If you want to avoid this step, pass proper
        ``init_params`` keyword argument to estimator's constructor.

        Parameters
        ----------
        obs : list
            List of array-like observation sequences, each of which
            has shape (n_i, n_features), where n_i is the length of
            the i_th observation.

        Notes
        -----
        In general, `logprob` should be non-decreasing unless
        aggressive pruning is used.  Decreasing `logprob` is generally
        a sign of overfitting (e.g. a covariance parameter getting too
        small).  You can fix this by getting more training data,
        or strengthening the appropriate subclass-specific regularization
        parameter.
        """
        self._init(obs, self.init_params)

        logprob = []
        for i in range(self.n_iter):
            # Expectation step
            stats = self._initialize_sufficient_statistics()
            curr_logprob = 0
            for seq in obs:
                framelogprob = self._compute_log_likelihood(seq)
                lpr, fwdlattice = self._do_forward_pass(framelogprob)
                bwdlattice = self._do_backward_pass(framelogprob)
                gamma = fwdlattice + bwdlattice
                posteriors = np.exp(gamma.T - logsumexp(gamma, axis=1)).T
                curr_logprob += lpr
                self._accumulate_sufficient_statistics(
                    stats, seq, framelogprob, posteriors, fwdlattice,
                    bwdlattice, self.params)
            logprob.append(curr_logprob)

            # Check for convergence.
            if i > 0 and logprob[-1] - logprob[-2] < self.thresh:
                break

            # Maximization step
            self._do_mstep(stats, self.params, three_para)
        #print("Logprob of all M-steps: %s" %logprob, file=sys.stderr)
        self.em_prob = logprob[-1]
        return self
    
    def _update_distr(self, mu, alpha):
        """Update distributions assigned to each state with new mu and alpha"""
        raw1 = [NegBin(mu[0, 0], alpha[0, 0]), NegBin(mu[0, 1], alpha[0, 1]), NegBin(mu[0, 2], alpha[0, 2])]
        raw2 = [NegBin(mu[1, 0], alpha[1, 0]), NegBin(mu[1, 1], alpha[1, 1]), NegBin(mu[1, 2], alpha[1, 2])]
        
        self.neg_distr = np.matrix([raw1, raw2]) #matrix of all Neg. Bin. Distributions, columns=HMM's state (3), row=#samples (2)
        
    def get_alpha(self, m):
        """Return alpha for a given mu based on empirical variance"""
        var = self.func(m)
        try:
            return max((var - m) / m**2, 1e-300)
        except Warning:
            if m**2 > 1e-300:
                return max((var - m) / m**2, 1e-300)
            else:
                return 1e-300
    
    def _compute_log_likelihood(self, X):
        matrix = []
        lookup = {}
        for x in X: #over all observations
            row = []
            for i in range(self.n_components): #over number of HMM's state
                r_sum = 0
                for j in range(self.n_features): #over dim
                    it = range(self.dim[0]) if j == 0 else range(self.dim[0], self.dim[0] + self.dim[1]) #grab proper ob
                    for k in it:
                        index = (int(x[k]), i, j)
                        if lookup.has_key( index ):
                            r_sum += lookup[index]
                        else:
                            y = float(self.neg_distr[j,i].logpdf(x[k]))
                            lookup[index] = y
                            r_sum += y
                row.append(r_sum)
        
            matrix.append(row)
        return np.asarray(matrix)
    
    
    def _generate_sample_from_state(self, state, random_state=None):
        output = []
        for i, d in enumerate(self.dim):
            for _ in range(d):
                output.append( self.neg_distr[i,state].rvs() )
        
        return np.array(output)
    
    def _initialize_sufficient_statistics(self):
        stats = super(NegBinRepHMM, self)._initialize_sufficient_statistics()
        stats['post'] = np.zeros([self.n_features, self.n_components])
        stats['post_emission'] = np.zeros([self.n_features, self.n_components]) #dim X states
        
        return stats
    
    def _help_accumulate_sufficient_statistics(self, obs, stats, posteriors):
        for t, symbol in enumerate(obs):
            stats['post'][0] += posteriors[t]
            stats['post'][1] += posteriors[t]
            
            pot_it = [range(self.dim[0]), range(self.dim[0], self.dim[0] + self.dim[1])] #consider both classes
            for j, it in enumerate(pot_it):
                for i in it:
                    stats['post_emission'][j] += posteriors[t] * symbol[i]
        
        stats['post'][0] = stats['post'][0] * self.dim[0]
        stats['post'][1] = stats['post'][1] * self.dim[1]
        
        stats['posterior'] = np.copy(posteriors)
    
    def _valid_posteriors(self, posteriors, obs):
    
        warnings.filterwarnings('error')
        
        for i in range(len(obs)):
            state_1 = False
            c1, c2 = np.mean(obs[i][:self.dim[0]]), np.mean(obs[i][self.dim[0]:]) #counts of samples
    
            if posteriors[i][0] > 0.5: 
                state_1 = True
                
            if c1 > c2: #state 1
                if fabs(posteriors[i][2] - 1) < 1e-200:
                    posteriors[i] = np.array([1, 0, 0])
                else:
                    if not state_1 and posteriors[i][2] > posteriors[i][1]:
                        try:                        
                            post_s2 = 0
                            post_s0 = posteriors[i][0] / (posteriors[i][0] + posteriors[i][1])
                            post_s1 = posteriors[i][1] / (posteriors[i][0] + posteriors[i][1])
                            posteriors[i] = np.array([post_s0, post_s1, post_s2])
                        except RuntimeWarning:
                            print(posteriors[i], c1, c2, file=sys.stderr)
            
            if c2 > c1: #state 2
                if fabs(posteriors[i][1] - 1) < 1e-200:
                    posteriors[i] = np.array([1, 0, 0])
                else:
                    if not state_1 and posteriors[i][1] > posteriors[i][2]:
                        try:
                            post_s1 = 0
                            post_s0 = posteriors[i][0] / (posteriors[i][0] + posteriors[i][2])
                            post_s2 = posteriors[i][2] / (posteriors[i][0] + posteriors[i][2])
                            posteriors[i] = np.array([post_s0, post_s1, post_s2])
                        except RuntimeWarning:
                            print(posteriors[i], c1, c2, file=sys.stderr)
        
        return posteriors

    def _accumulate_sufficient_statistics(self, stats, obs, framelogprob,
                                      posteriors, fwdlattice, bwdlattice,
                                      params):
        super(NegBinRepHMM, self)._accumulate_sufficient_statistics(
            stats, obs, framelogprob, posteriors, fwdlattice, bwdlattice,
            params)
        posteriors = self._valid_posteriors(posteriors, obs)
        self._help_accumulate_sufficient_statistics(obs, stats, posteriors)        
    
    def _do_mstep(self, stats, params, three_para):
        super(NegBinRepHMM, self)._do_mstep(stats, params)
        
        if three_para:
            self.mu[0,1] = (stats['post_emission'][0][1] + stats['post_emission'][1][2]) / (stats['post'][0][1] + stats['post'][1][2])
            self.mu[1,1] = (stats['post_emission'][1][1] + stats['post_emission'][0][2]) / (stats['post'][1][1] + stats['post'][0][2])
            self.mu[0,0] = (stats['post_emission'][0][0] + stats['post_emission'][1][0]) / (stats['post'][0][0] + stats['post'][1][0])
            
            self.mu[1,2] = self.mu[0,1]
            self.mu[0,2] = self.mu[1,1]
            self.mu[1,0] = self.mu[0,0]
        else:
            self.mu[0,1] = (stats['post_emission'][0][1] + stats['post_emission'][1][2]) / (stats['post'][0][1] + stats['post'][1][2])
            self.mu[1,1] = (stats['post_emission'][1][1] + stats['post_emission'][0][2] + stats['post_emission'][0][0] + stats['post_emission'][0][1]) / (stats['post'][1][1] + stats['post'][0][2] + stats['post'][0][0] + stats['post'][0][1])
            
            self.mu[0,0] = self.mu[1,1]
            self.mu[1,2] = self.mu[0,1]
            self.mu[0,2] = self.mu[1,1]
            self.mu[1,0] = self.mu[0,0]
        
        self.alpha = np.matrix([map(lambda m: self.get_alpha(m), np.asarray(self.mu[i])[0]) for i in range(self.n_features)])
        self._update_distr(self.mu, self.alpha)
       
    def merge_distr(self):
        f = self.count_s2 / float(self.count_s1 + self.count_s2) #TODO exp_data.
        
        for el in [self.mu, self.alpha]:
            high = min(el[0,1], el[1,2]) + f * fabs(el[0,1] - el[1,2])
            low = min(el[1,1], el[0,2]) + f * fabs(el[1,1] - el[0,2])
            med = np.mean([el[0,0], el[1,0]])
            el[0,1] = high
            el[1,2] = high
            el[1,1] = low
            el[0,2] = low
            el[0,0] = low #min(med, low)
            el[1,0] = low #min(med, low)
        
        self._update_distr(self.mu, self.alpha)

    
if __name__ == '__main__':
    alpha = np.matrix([[0.2, 0.2, 0.2], [0.2, 0.2, 0.2]])
    mu = np.matrix([[15.,100.,10.], [10.,10.,100.]])
    f = lambda x: 0.4*x**2 + 1
    
    dim_cond_1 = 2
    dim_cond_2 = 3

    m = NegBinRepHMM(alpha = alpha, mu = mu, dim_cond_1 = dim_cond_1, dim_cond_2 = dim_cond_2, func=f)
    
    X, Z = m.sample(20)
    
    m2 = NegBinRepHMM(alpha = alpha, mu = np.matrix([[50.,130.,110.], [60.,100.,120.]]), dim_cond_1 = dim_cond_1, dim_cond_2 = dim_cond_2, func=f)
    m2.fit([X], three_para=False)
    
    posteriors = m.predict_proba(X)
    e = m2.predict(X)
    for i, el in enumerate(X):
        print(el, Z[i], e[i], Z[i] == e[i], sep='\t', file=sys.stderr)
    #print(np.max(posteriors, axis=1))
    print(m2.mu)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2014-2016 Manuel Allhoff (allhoff@aices.rwth-aachen.de)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

%prog <BED,BAM> <BED,BAM> <NAME> <STEP_WIDTH> <Zero-counts 1/0> <hg19/mm9>

Computes p-q list, k and a from Diaz et al., 2012.

Features must have same length in both input files.

@author: Manuel Allhoff
"""

from __future__ import print_function
from optparse import OptionParser
from HTSeq import GenomicPosition, GenomicArray, GenomicInterval
from math import fabs
import pysam, sys, operator, os.path
from itertools import chain


class HelpfulOptionParser(OptionParser):
    """An OptionParser that prints full help on errors."""

    def error(self, msg):
        self.print_help(sys.stderr)
        self.exit(2, "\n%s: error: %s\n" % (self.get_prog_name(), msg))


def _accumulate(iterable):
    """Return running totals"""
    it = iter(iterable)
    total = next(it)
    yield [total[0], total[1]]
    for e1, e2 in it:
        total[0] = total[0] + e1
        total[1] = total[1] + e2
        yield [total[0], total[1]]


def get_feature_len(path):
    """Return length of reads"""
    filename, fileextension = os.path.splitext(path)

    if fileextension == '.bed':
        with open(path, 'r') as f:
            for line in f:
                tmp = line.split('\t')
                return int(tmp[2]) - int(tmp[1])
    elif fileextension == '.bam':
        samfile = pysam.Samfile(path, "rb")
        for read in samfile.fetch():
            if not read.is_unmapped:
                return read.alen


def _get_read_info(path):
    """Yield tuple (chr, pos) for different fileformats."""
    filename, fileextension = os.path.splitext(path)

    if fileextension == '.bed':
        with open(path, 'r') as f:
            for line in f:
                tmp = line.split('\t')
                yield tmp[0], int(tmp[1])
    elif fileextension == '.bam':
        samfile = pysam.Samfile(path, "rb")
        for read in samfile.fetch():
            if not read.is_unmapped:
                yield samfile.getrname(read.tid), read.pos


def get_count_list(CHROM_LEN, path, stop=False):
    """Compute list of read's starting positions based on HTSeq's genomic array."""
    genomic_array = GenomicArray(CHROM_LEN, stranded=False, typecode='i', storage='step')
    i = 0
    chromosomes = set()
    for chrom, pos in _get_read_info(path):
        i += 1

        if stop and i == 5000000:
            break

        if not chrom.startswith("chr"):
            chrom = "chr" + chrom

        if chrom not in chromosomes:
            chromosomes.add(chrom)

        # ignore reads that fall out of chromosome borders, should not be necassary!
        if chrom not in CHROM_LEN.keys() or pos >= CHROM_LEN[chrom] or pos < 0:
            # print("illegal read on chromosome %s at positions %s not"%(chrom, pos), file=sys.stderr)
            continue

        genomic_array[GenomicPosition(chrom, pos)] += 1

    # print('count_list', i, file=sys.stderr)

    return genomic_array, chromosomes


def _get_overrun(chrom, i, end, step_width, count_list, feature_len):
    """Return overrun of reads that fall in two bins"""
    help_c1 = filter(lambda x: x[0].start + feature_len > end and x[1] is not 0,
                     list(count_list[GenomicInterval(chrom, i, end)].steps()))
    overrun = 0 if not help_c1 else sum(map(lambda x: x[1], help_c1))

    return overrun


def write_pq_list(pq_list, max_index, max_value, factor1, factor2, filename):
    """Write p,q-list to file"""
    if pq_list:
        with open(filename, 'w') as f:
            print('#max index', 'max value', 'factor1', 'factor2', sep='\t', file=f)
            print('#' + str(max_index), str(max_value), str(factor1), str(factor2), sep='\t', file=f)
            for p, q in pq_list:
                print(p, q, file=f)


def get_bins(chrom_len, chromosomes, count_list, step_width, feature_len):
    """Creates list of bins of length <step_width> with values describing 
    the number of reads that fall into a bin.
    <count_list> has to be created with 'get_count_list' 
    It returns a dict like: {'chr1' [0,10,2,0,...], 'chr2' ...} 
    where the first list entry gives the first bin and so on."""
    result = {}
    for chrom in chromosomes:
        overrun = 0
        if not chrom_len.has_key(chrom):
            #             print("Warning: %s not found, do not consider" %chrom, file=sys.stderr)
            pass
        else:
            # print("... considering %s..."%chrom, file=sys.stderr)
            for i in range(0, chrom_len[chrom], step_width):
                end = min(i + step_width, chrom_len[chrom])
                counts = reduce(lambda x, y: x + y, count_list[GenomicInterval(chrom, i, end)])
                count_list[GenomicInterval(chrom, i, end)] = 0
                counts += overrun

                if chrom in result.keys():
                    result[chrom].append(counts)
                else:
                    result[chrom] = [counts]

                overrun = _get_overrun(chrom, i, end, step_width, count_list, feature_len)

    return result


def _get_lists(count_list, zero_counts, two_sample=False):
    if two_sample:
        count_list.sort(key=lambda x: x[0] + x[1])
        if not zero_counts:
            count_list = filter(lambda x: x[0] != 0.0 or x[1] != 0.0, count_list)
        pre_pq_list = list(_accumulate(count_list))
        pq_list = pre_pq_list
        max_index = int(len(pq_list) * 0.5)
        max_value = fabs(pq_list[max_index][1] - pq_list[max_index][0])
    else:
        count_list.sort(key=lambda a: a[0])
        if not zero_counts:
            count_list = filter(lambda x: x[0] != 0.0, count_list)
        pre_pq_list = list(_accumulate(count_list))
        # get k, a from Diaz et al., 2012, compute p, q from Diaz et al., 2012
        pq_list = map(lambda x: [x[0] / float(pre_pq_list[-1][0]), x[1] / float(pre_pq_list[-1][1])], pre_pq_list)
        max_index, max_value = max(enumerate(map(lambda x: fabs(x[0] - x[1]), pq_list)), key=operator.itemgetter(1))

    return pq_list, max_index, max_value, \
           float(pq_list[max_index][0]) / pq_list[max_index][1], \
           float(pq_list[max_index][1]) / pq_list[max_index][0]


def get_binstats(chrom_len, count_list_1, count_list_2, feature_len, chromosomes, step_width=1000, zero_counts=True,
                 two_sample=False):
    """Compute p,q-list from Diaz et al., 2012 with k, a and coressponding factors
    for normalization.
    Also give list of tuples (x,y) describing bins where x,y are the values for each bin."""
    # create dict of counts
    # print("Dividing genome into bins...", file=sys.stderr)
    if set(chrom_len.keys()) & chromosomes == set():  # if chrom in input does not overlap chrom in IP channel
        return None, None, None, None, None

    counts_dict_2 = get_bins(chrom_len, chromosomes, count_list_2, step_width, feature_len)
    counts_dict_1 = get_bins(chrom_len, chromosomes, count_list_1, step_width, feature_len)
    # merge values with zip, obtain [ [(),()...], [(),(),..] ]
    tmp_list = [zip(counts_dict_1[chrom], counts_dict_2[chrom]) for chrom in counts_dict_2.keys()]
    # convert tmp_list to [[],[],[]...]
    count_list = map(lambda x: list(x), list(chain.from_iterable(tmp_list)))

    return _get_lists(count_list, zero_counts, two_sample)


def work(first_path, second_path, step_width, zero_counts, two_sample, chrom_sizes_dict, stop):
    """work"""
    CHROM_LEN = chrom_sizes_dict  # CHROM_LEN_HUMAN if genome == 'hg19' else CHROM_LEN_MOUSE
    # counts_1, counts_1 is a genomicarray (HTSeq) describing chr und pos of reads
    # print("Reading first input file...", file=sys.stderr)
    counts_1, chromosomes1 = get_count_list(CHROM_LEN, first_path, stop=stop)

    # print("Reading second input file...", file=sys.stderr)
    counts_2, chromosomes2 = get_count_list(CHROM_LEN, second_path, stop=stop)
    #     print("...done", file=sys.stderr)
    chromosomes = chromosomes1 | chromosomes2

    # count_list1, count_list2 are dict like {'chr1': [10,2,0...], ...} describing the bins
    pq_list, max_index, max_value, factor1, factor2 \
        = get_binstats(CHROM_LEN, counts_1, counts_2, \
                       get_feature_len(first_path), chromosomes, step_width, zero_counts, two_sample)

    return pq_list, max_index, max_value, factor1, factor2, chromosomes


def get_normalization_factor(first_path, second_path, step_width, zero_counts, filename, debug, chrom_sizes_dict,
                             two_sample=False, stop=False):
    """Return normalization factor (see Diaz et al) for the input
    if two_sample is True: compare sample with index of 0.15"""
    # take two largest chromosomes for analysis:
    # tmp = chrom_sizes_dict.items()
    # tmp.sort(key=lambda x: x[1], reverse=True)
    # tmp = dict(tmp[:min(len(tmp), 5)])
    tmp = {}
    for k in chrom_sizes_dict.keys():
        new_k = k if k.startswith('chr') else 'chr' + k
        tmp[new_k] = chrom_sizes_dict[k]
    # tmp = chrom_sizes_dict
    # print("For input normalization consider chromosomes: %s" %(", ".join(tmp.keys())), file=sys.stderr)

    pq_list, max_index, max_value, factor1, factor2, chromosomes = \
        work(first_path, second_path, step_width, zero_counts, two_sample, tmp, stop)

    if debug:
        write_pq_list(pq_list, max_index, max_value, factor1, factor2, filename + '-pqlist')

    # print("norm sums 1 : ", sum([i for (i,j) in pq_list]), file=sys.stderr)
    # print("norm sums 2 : ", sum([j for (i,j) in pq_list]), file=sys.stderr)

    if pq_list == None:
        return None, None

    if two_sample:
        l = 0.5
        s1 = sum([i for (i, j) in pq_list[:int(len(pq_list) * l)]])
        s2 = sum([j for (i, j) in pq_list[:int(len(pq_list) * l)]])

        # print("norm sums half: ", s1, s2, file=sys.stderr)

        if s1 > s2:
            return 2, factor1
        else:
            return 1, factor1
    else:
        return -1, factor1



# Python Libraries
from __future__ import print_function
import os
import sys
import time
import shutil
import natsort
import datetime
from collections import *

# Local Libraries
import numpy
numpy.seterr(divide='ignore', invalid='ignore')

from scipy import stats
import matplotlib.pyplot as plt
import matplotlib.patches as patches
import scipy.cluster.hierarchy as sch
from matplotlib.backends.backend_pdf import PdfPages
from matplotlib.ticker import FuncFormatter

# Distal Libraries
from rgt.GeneSet import GeneSet
from rgt.SequenceSet import SequenceSet
from rgt.AnnotationSet import AnnotationSet
from RNADNABindingSet import RNADNABindingSet
from rgt.GenomicRegionSet import GenomicRegionSet
from rgt.motifanalysis.Statistics import multiple_test_correction
from rgt.Util import SequenceType, Html, GenomeData, OverlapType
from triplexTools import dump, load_dump, print2, get_rna_region_str, connect_rna,\
    get_sequence, run_triplexator, dbd_regions, lineplot, value2str, rank_array,\
    split_gene_name, region_link_internet


# Color code for all analysis
target_color = "mediumblue"
nontarget_color = "darkgrey"
sig_color = "powderblue"


class PromoterTest:
    """Test the association between given triplex and differential expression genes"""
    def __init__(self, gene_list_file, bed, bg, organism, promoterLength, rna_name,
                 summary, temp, output, showdbs=None, score=False, scoreh=False,
                 filter_havana=True, protein_coding=False, known_only=True, gtf=None):
        """Initiation"""
        self.organism = organism
        if gtf: self.gtf = gtf
        else: self.gtf = organism
        genome = GenomeData(organism)
        self.rna_name = rna_name
        self.genome_path = genome.get_genome()
        self.showdbs = showdbs
        self.scores = None
        self.scoreh = scoreh
        self.motif = OrderedDict()
        self.sig_DBD = []
        self.stat = OrderedDict(name=rna_name, genome=self.gtf)
        ####################################################################################
        # Input BED files
        if bed and bg:
            self.de_regions = GenomicRegionSet("de genes")
            self.nde_regions = GenomicRegionSet("nde genes")
            self.de_regions.read_bed(bed)
            self.de_regions.remove_duplicates()

            # score
            if score:
                self.scores = self.de_regions.get_score_dict()
                # for promoter in self.de_regions:
                #     if promoter.data: self.scores.append(promoter.data.split("\t")[0])
                #     else: self.scores = None
                # self.score_dict = self.de_regions.get_score_dict()

            # try: self.de_regions = self.de_regions.gene_association(organism=self.organism)
            # except: pass
            self.nde_regions.read_bed(bg)
            # try: self.nde_regions = self.nde_regions.gene_association(organism=self.organism)
            # except: pass
            self.nde_regions.remove_duplicates()

        ####################################################################################
        # Input DE gene list
        else:
            if gtf:
                g = os.path.basename(gtf)
                dumpname = "_".join(["dump", gene_list_file.rpartition("/")[-1].rpartition(".")[0],
                                     g.rpartition(".")[0],
                                     filter_havana + protein_coding + known_only])
            else:
                dumpname = "_".join(["dump", gene_list_file.rpartition("/")[-1].rpartition(".")[0],
                                     filter_havana + protein_coding + known_only])
            try:
                data = load_dump(path=temp, filename=dumpname)
                self.de_gene = data[0]
                self.ensembl2symbol = data[1]
                self.nde_gene = data[2]
                self.de_regions = data[3]
                self.nde_regions = data[4]
                if score: self.scores = data[5]

            except:

                t1 = time.time()
                ######################################################
                # DE gene regions
                self.de_gene = GeneSet("de genes")

                if score:
                    self.de_gene.read(gene_list_file, score=True)
                    # self.de_gene.read_expression(geneListFile=gene_list_file, header=scoreh, valuestr=True)
                else:
                    self.de_gene.read(gene_list_file)
                # When there is no genes in the list
                if len(self.de_gene) == 0:
                    print("Error: No genes are loaded from: " + gene_list_file)
                    print("Please check the format.")
                    try:
                        shutil.rmtree(output)
                    except:
                        pass
                    sys.exit(1)

                print2(summary, "   \t" + str(len(self.de_gene)) + "\tinput genes ")

                if "REFSEQ" not in self.gtf.upper():

                    # Setting annotationSet
                    if filter_havana== "T":
                        filter_havana = True
                    else:
                        filter_havana = False
                    if protein_coding == "T":
                        protein_coding = True
                    else:
                        protein_coding = False
                    if known_only == "T":
                        known_only = True
                    else:
                        known_only = False

                    ann = AnnotationSet(gene_source=self.gtf, alias_source=organism,
                                        filter_havana=filter_havana,
                                        protein_coding=protein_coding,
                                        known_only=known_only)

                    t2 = time.time()
                    print("\t" + str(datetime.timedelta(seconds=round(t2 - t1))))


                    # Generate a dict for ID transfer
                    de_ensembl, unmap_gs, self.ensembl2symbol = ann.fix_gene_names(gene_set=self.de_gene, output_dict=True)
                    print2(summary, "   \t" + str(len(de_ensembl)) + "\tgenes are mapped to Ensembl ID")
                    if len(unmap_gs) > 0:
                        print2(summary, "   \t" + str(len(unmap_gs)) + "\tunmapped genes: " + ",".join(unmap_gs))

                    self.de_gene.genes = de_ensembl
                    de_prom, unmapped_gene_list = ann.get_promoters(promoterLength=promoterLength,
                                                                    gene_set=self.de_gene,
                                                                    unmaplist=True, variants=False)
                    print2(summary, "   \t" + str(len(de_prom)) + "\tmapped promoters")
                    if len(unmapped_gene_list) > 0:
                        print2(summary, "   \t" + str(len(unmapped_gene_list)) + "\tunmapped promoters: " + ",".join(
                            unmapped_gene_list))

                    ######################################################
                    # NonDE gene regions

                    nde_ensembl = [g for g in ann.symbol_dict.keys() if g not in de_ensembl]
                    print2(summary, "   \t" + str(len(nde_ensembl)) + "\tnon-target genes")
                    self.nde_gene = GeneSet("nde genes")
                    self.nde_gene.genes = nde_ensembl

                    # Get promoters from nonDE gene
                    nde_prom, unmapped_gene_list = ann.get_promoters(promoterLength=promoterLength,
                                                                     gene_set=self.nde_gene,
                                                                     unmaplist=True)
                    print2(summary, "   \t" + str(len(nde_prom)) + "\tmapped non-target promoters")

                    if score:

                        self.scores = {}
                        for p in de_prom:
                            # d = c.split("\t")
                            p.data = self.de_gene.values[self.ensembl2symbol[p.name]]
                        self.scores = de_prom.get_score_dict()
                        # print(self.scores.keys()[:5])
                        # print(self.scores.values()[:5])

                else:
                ## Refseq
                    print("   \t" + "Loading from RefSeq annotation")
                    genes = GenomicRegionSet("genes")
                    genes.read_bed(genome.get_genes_refseq())
                    de_genes = genes.by_names(self.de_gene, load_score=score)
                    self.de_gene = de_genes.get_GeneSet()
                    de_prom = de_genes.get_promoters(length=promoterLength)
                    de_prom.merge(namedistinct=True, strand_specific=True)
                    nde_genes = genes.by_names(self.de_gene, load_score=score, background=True)

                    self.nde_gene = nde_genes.get_GeneSet()
                    nde_prom = nde_genes.get_promoters(length=promoterLength)
                    nde_genes.merge(namedistinct=True, strand_specific=True)
                    self.ensembl2symbol = None
                    if score:
                        self.scores = {}
                        for p in de_prom:
                            p.data = str(self.de_gene.values[p.name])
                        self.scores = de_prom.get_score_dict()
                        # self.scores = de_genes.get_score_dict()
                        print(self.scores.keys()[:5])
                        print(self.scores.values()[:5])

                # de_prom.merge(namedistinct=True, strand_specific=True)
                print2(summary, "   \t" + str(len(de_prom)) + "\tmerged promoters ")

                print2(summary, "   \t" + str(len(de_prom)) + "\tunique target promoters are loaded")

                # for promoter in de_prom:
                #     # if self.ensembl2symbol:
                #     try:
                #         gene_sym = self.ensembl2symbol[promoter.name]
                #     except:
                #         gene_sym = promoter.name
                #
                #     if score:
                #         try:
                #             try:
                #                 s = self.de_gene.values[promoter.name]
                #             except:
                #                 s = self.de_gene.values[gene_sym.upper()]
                #         except:
                #             try:
                #                 print("Warning: " + promoter.name + "\tcannot be mapped to get its score.")
                #             except:
                #                 print("Warning: " + gene_sym + "\tcannot be mapped to get its score.")
                #             s = 0
                #         self.scores[promoter.toString()] = s

                self.de_regions = de_prom

                nde_prom.merge(namedistinct=True)
                print2(summary, "   \t" + str(len(nde_prom)) + "\tmerged non-target promoters")

                self.nde_regions = nde_prom

                print2(summary, "   \t" + str(len(nde_prom)) + "\tunique non-target promoters are loaded")
                # Loading score

                data = [self.de_gene, self.ensembl2symbol, self.nde_gene, self.de_regions, self.nde_regions]
                if score: data.append(self.scores)
                dump(object=data, path=temp, filename=dumpname)

            self.stat["target_regions"] = str(len(self.de_regions))
            self.stat["background_regions"] = str(len(self.nde_regions))

    def get_rna_region_str(self, rna, expfile):
        """Getting the rna region from the information header with the pattern:
                REGION_chr3_51978050_51983935_-_
            or  chr3:51978050-51983935 -    """
        self.rna_regions = get_rna_region_str(rna)

        if self.rna_regions and len(self.rna_regions[0]) == 5:
            self.rna_expression = float(self.rna_regions[0][-1])
        elif expfile:
            with open(expfile) as f:
                for line in f:
                    l = line.strip().split()
                    if self.rna_name == l[0].partition(".")[0]:
                        self.rna_expression = l[1]
        else:
            self.rna_expression = "n.a."

    def connect_rna(self, rna, temp):
        d = connect_rna(rna, temp, self.rna_name)
        self.stat["exons"] = str(d[0])
        self.stat["seq_length"] = str(d[1])

    def search_triplex(self, temp, l, e, c, fr, fm, of, mf, par, remove_temp=False):
        print("    \tRunning Triplexator...")
        rna = os.path.join(temp, "rna_temp.fa")
        self.triplexator_p = [l, e, c, fr, fm, of, mf]

        # DE
        get_sequence(dir=temp, filename=os.path.join(temp, "de.fa"), regions=self.de_regions,
                     genome_path=self.genome_path)
        run_triplexator(ss=rna, ds=os.path.join(temp, "de.fa"),
                        output=os.path.join(temp, "de.txp"),
                        l=l, e=e, c=c, fr=fr, fm=fm, of=of, mf=mf, par=par)

        # non-DE
        get_sequence(dir=temp, filename=os.path.join(temp, "nde.fa"), regions=self.nde_regions,
                     genome_path=self.genome_path)
        run_triplexator(ss=rna, ds=os.path.join(temp, "nde.fa"),
                        output=os.path.join(temp, "nde.txp"),
                        l=l, e=e, c=c, fr=fr, fm=fm, of=of, mf=mf, par=par)

        # os.remove(os.path.join(args.o,"rna_temp.fa"))
        if remove_temp:
            os.remove(os.path.join(temp, "de.fa"))
            os.remove(os.path.join(temp, "nde.fa"))

    def count_frequency(self, temp, remove_temp, cutoff, l, obedp=False):
        """Count the frequency between DE genes and non-DE genes with the given BindingSiteSet"""

        len_de = len(self.de_regions)
        len_nde = len(self.nde_regions)
        # print([len_de, len_nde])

        self.frequency = {}
        self.frequency["promoters"] = {"de": OrderedDict(), "nde": OrderedDict()}

        ########################################################
        # Count the number of targeted promoters on each merged DNA Binding Domain
        print("\tCounting frequency of promoters on DBD...")
        self.txp_de = RNADNABindingSet("DE")

        self.txp_de.read_txp(os.path.join(temp, "de.txp"), dna_fine_posi=False)
        print("\t\t" + str(len(self.txp_de)) + "\tBinding de promoters")
        self.txp_de.remove_duplicates()
        self.txp_de.merge_rbs(rm_duplicate=True, cutoff=cutoff,
                              region_set=self.de_regions, name_replace=self.de_regions)

        self.rbss = self.txp_de.merged_dict.keys()
        for rbs in self.rbss:
            # DE
            self.txp_de.merged_dict[rbs].remove_duplicates()
            l1 = len(self.txp_de.merged_dict[rbs])
            self.frequency["promoters"]["de"][rbs] = [l1, len_de - l1]

        self.txp_nde = RNADNABindingSet("non-DE")
        self.txp_nde.read_txp(os.path.join(temp, "nde.txp"), dna_fine_posi=False)
        print("\t\t" + str(len(self.txp_nde)) + "\tBinding nde promoters")
        self.txp_nde.remove_duplicates()
        self.txp_nde.merge_rbs(rbss=self.rbss, region_set=self.nde_regions,
                               rm_duplicate=True)  # , asgene_organism=self.organism)

        for rbs in self.rbss:
            l2 = len(self.txp_nde.merged_dict[rbs])
            self.frequency["promoters"]["nde"][rbs] = [l2, len_nde - l2]

        self.stat["DBSs_target_all"] = str(len(self.txp_de))
        self.stat["DBSs_background_all"] = str(len(self.txp_nde))

        ########################################################
        # Count the number of hits on the promoters from each merged DBD

        print("\tCounting frequency of binding sites on DBD...")
        ####################
        # DE
        self.txp_def = RNADNABindingSet("DE")
        self.txp_def.read_txp(os.path.join(temp, "de.txp"), dna_fine_posi=True)
        self.txp_def.merge_rbs(rbss=self.rbss, rm_duplicate=True,
                               name_replace=self.de_regions)  # asgene_organism=self.organism
        print("\t\t" + str(len(self.txp_def)) + "\tBinding sites on de promoters")


        # Promoter profiling
        self.promoter = {"de": {},"nde": {}}

        self.promoter["de"]["rd"] = self.txp_def.sort_rd_by_regions(regionset=self.de_regions)
        # self.promoter["de"]["merged_dbs"] = {}
        self.promoter["de"]["dbs"] = {}
        self.promoter["de"]["dbs_coverage"] = {}

        for promoter in self.de_regions:
            dbs = self.promoter["de"]["rd"][promoter.toString()].get_dbs()
            m_dbs = dbs.merge(w_return=True)
            self.promoter["de"]["dbs"][promoter.toString()] = len(dbs)
            # self.promoter["de"]["merged_dbs"][promoter.toString()] = len(m_dbs)
            self.promoter["de"]["dbs_coverage"][promoter.toString()] = float(m_dbs.total_coverage()) / len(promoter)

        ######################
        # nDE
        self.promoter["nde"]["dbs"] = {}
        self.promoter["nde"]["dbs_coverage"] = {}

        # ndef_dbs = GenomicRegionSet("nde_dbs")

        self.txp_nde = RNADNABindingSet("non-DE")
        self.txp_nde.read_txp(os.path.join(temp, "nde.txp"), dna_fine_posi=False)
        self.txp_nde.merge_rbs(rbss=self.rbss, rm_duplicate=True)  # , asgene_organism=self.organism)


        ndef_dbs = self.txp_nde.get_dbs()

        counts = self.nde_regions.counts_per_region(regionset=ndef_dbs)

        coverage = self.nde_regions.coverage_per_region(regionset=ndef_dbs)
        for i, p in enumerate(self.de_regions):
            self.promoter["nde"]["dbs"][p.toString()] = counts[i]
            # self.promoter["nde"]["merged_dbs"][p.toString()] = mcounts[i]
            self.promoter["nde"]["dbs_coverage"][p.toString()] = coverage[i]

        if self.showdbs:
            self.frequency["hits"] = {"de": OrderedDict(), "nde": OrderedDict()}
            numdbs_def = len(self.txp_def.get_dbs(rm_duplicate=True))
            numdbs_ndef = len(self.txp_ndef.get_dbs(rm_duplicate=True))
            for rbs in self.rbss:
                # DE
                l1 = len(self.txp_def.merged_dict[rbs])
                self.frequency["hits"]["de"][rbs] = [l1, numdbs_def - l1]
                # non-DE
                l2 = len(self.txp_ndef.merged_dict[rbs])
                self.frequency["hits"]["nde"][rbs] = [l2, numdbs_ndef - l2]

        if remove_temp:
            os.remove(os.path.join(temp, "de.txp"))

        os.remove(os.path.join(temp, "nde.txp"))

        if obedp:
            try:
                output = self.de_regions.change_name_by_dict(convert_dict=self.ensembl2symbol)
            except:
                output = self.de_regions
            output.write_bed(filename=os.path.join(temp, obedp + "_target_promoters.bed"))

            self.txp_de.write_bed(filename=os.path.join(temp, obedp + "_target_promoters_dbs.bed"),
                                  associated=self.organism)
            self.txp_def.write_bed(filename=os.path.join(temp, obedp + "_dbss.bed"),
                                   remove_duplicates=False, associated=self.organism)

    def fisher_exact(self, alpha):
        """Return oddsratio and pvalue"""
        self.oddsratio = {}
        self.pvalue = {}
        pvalues = []

        for rbs in self.frequency["promoters"]["de"]:
            table = numpy.array([self.frequency["promoters"]["de"][rbs], self.frequency["promoters"]["nde"][rbs]])
            # print(table)
            self.oddsratio[rbs], p = stats.fisher_exact(table, alternative="greater")
            pvalues.append(p)

        # correction
        if len(self.frequency["promoters"]["de"].keys()) > 1:
            reject, pvals_corrected = multiple_test_correction(pvalues, alpha=alpha, method='indep')
        else:
            pvals_corrected = pvalues
        for i, rbs in enumerate(self.frequency["promoters"]["de"].keys()):
            self.pvalue[rbs] = pvals_corrected[i]
            if pvals_corrected[i] < alpha:
                self.sig_DBD.append(rbs)

        try: self.stat["p_value"] = str(min(pvals_corrected))
        except: self.stat["p_value"] = "1"

        if self.showdbs:
            self.hoddsratio = {}
            self.hpvalue = {}
            pvalues = []
            for rbs in self.frequency["hits"]["de"]:
                table = numpy.array([self.frequency["hits"]["de"][rbs], self.frequency["hits"]["nde"][rbs]])
                self.hoddsratio[rbs], p = stats.fisher_exact(table, alternative="greater")
                pvalues.append(p)
            # correction
            if len(self.frequency["hits"]["de"].keys()) > 1:
                reject, pvals_corrected = multiple_test_correction(pvalues, alpha=alpha, method='indep')
            else:
                pvals_corrected = pvalues
            for i, rbs in enumerate(self.frequency["hits"]["de"].keys()):
                self.hpvalue[rbs] = pvals_corrected[i]
                if pvals_corrected[i] < alpha:
                    self.sig_DBD.append(rbs)

        # Generate a dict for DBS from sig. DBD

    def dbd_regions(self, output):
        dbd_regions(exons=self.rna_regions, sig_region=self.sig_DBD,
                    rna_name=self.rna_name, output=output)
        # if len(self.rna_regions[0]) == 4:
        #     self.stat["expression"] = str(self.rna_regions[0][3])
        self.stat["DBD_all"] = str(len(self.rbss))
        self.stat["DBD_sig"] = str(len(self.sig_DBD))

        sigDBD = GenomicRegionSet("DBD_sig")
        sigDBD.sequences = self.sig_DBD
        rbss = self.txp_de.get_rbs()
        overlaps = rbss.intersect(y=sigDBD, mode=OverlapType.ORIGINAL)
        self.stat["DBSs_target_DBD_sig"] = str(len(overlaps))
        rbss = self.txp_nde.get_rbs()
        overlaps = rbss.intersect(y=sigDBD, mode=OverlapType.ORIGINAL)
        self.stat["DBSs_background_DBD_sig"] = str(len(overlaps))

    def plot_lines(self, txp, rna, dirp, cut_off, log, ylabel, linelabel, filename, sig_region, ac=None, showpa=False):
        """Generate the plots for demonstration of RBS

        rna:     File path to the RNA sequence in FASTA format
        dir:     Directory where the figure is saved
        ac:      RNA accessibility data.
        """
        rnas = SequenceSet(name="rna", seq_type=SequenceType.RNA)
        rnas.read_fasta(rna)
        if not self.rna_name: self.rna_name = rnas[0].name

        self.rna_len = rnas.total_len()

        lineplot(txp=txp, rnalen=self.rna_len, rnaname=self.rna_name, dirp=dirp, sig_region=sig_region,
                 cut_off=cut_off, log=log, ylabel=ylabel, linelabel=linelabel,
                 filename=filename, ac=ac, showpa=showpa, exons=self.rna_regions)

    def barplot(self, dirp, filename, sig_region, dbs=False):
        """Generate the barplot to show the difference between target promoters and non-target promoters"""

        def to_percent(y, position):
            # Ignore the passed in position. This has the effect of scaling the default
            # tick locations.
            s = str(100 * y)
            return s + '%'

        f, ax = plt.subplots(1, 1, dpi=300, figsize=(6, 4))
        ind = range(len(self.rbss))
        width = 0.35

        if not dbs:
            propor_de = [float(b[0]) / len(self.de_regions) for b in self.frequency["promoters"]["de"].values()]
            propor_nde = [float(b[0]) / len(self.nde_regions) for b in self.frequency["promoters"]["nde"].values()]
        else:
            propor_de = [float(b[0]) / (b[0] + b[1]) for b in self.frequency["hits"]["de"].values()]
            propor_nde = [float(b[0]) / (b[0] + b[1]) for b in self.frequency["hits"]["nde"].values()]

        max_y = max([max(propor_de), max(propor_nde)]) * 1.2

        # Plotting
        rect = patches.Rectangle(xy=(1, 0), width=0.8, height=max_y, facecolor=sig_color,
                                 edgecolor="none", alpha=0.5, lw=None, label="Significant DBD")
        for i, rbs in enumerate(self.rbss):
            if rbs in sig_region:
                rect = patches.Rectangle(xy=(i + 0.05, 0), width=0.9, height=max_y, facecolor=sig_color,
                                         edgecolor="none", alpha=0.5, lw=None, label="Significant DBD")
                ax.add_patch(rect)

        rects_de = ax.bar([i + 0.15 for i in ind], propor_de, width, color=target_color,
                          edgecolor="none", label="Target promoters")
        rects_nde = ax.bar([i + 0.15 + width for i in ind], propor_nde, width, color=nontarget_color,
                           edgecolor="none", label="Non-target promoters")

        # Legend
        tr_legend, = plt.plot([1, 1], color=target_color, linewidth=6, alpha=1)
        ntr_legend, = plt.plot([1, 1], color=nontarget_color, linewidth=6, alpha=1)
        ax.legend([tr_legend, ntr_legend, rect], ["Target promoters", "Non-target promoters", "Significant DBD"],
                  bbox_to_anchor=(0., 1.02, 1., .102), loc=2, mode="expand", borderaxespad=0.,
                  prop={'size': 9}, ncol=3)
        tr_legend.set_visible(False)
        ntr_legend.set_visible(False)

        tick_size = 8
        # Y axis
        ax.set_ylim([0, max_y])
        formatter = FuncFormatter(to_percent)
        # Set the formatter
        ax.yaxis.set_major_formatter(formatter)
        ax.tick_params(axis='y', which='both', left='on', right='off', labelbottom='off')
        for tick in ax.yaxis.get_major_ticks(): tick.label.set_fontsize(9)
        ax.set_ylabel("Proportion of promoters (%)", fontsize=9, rotation=90)

        # X axis
        ax.set_xlim([0, len(self.rbss)])
        ax.tick_params(axis='x', which='both', bottom='off', top='off', labelbottom='on')
        ax.set_xticks([i + 0.5 for i in range(len(self.rbss))])
        ax.set_xticklabels([dbd.str_rna(pa=False) for dbd in self.rbss], rotation=35,
                           ha="right", fontsize=tick_size)
        for spine in ['top', 'right']:
            ax.spines[spine].set_visible(False)

        for tick in ax.xaxis.get_major_ticks(): tick.label.set_fontsize(9)
        ax.set_xlabel(self.rna_name + " DNA Binding Domains", fontsize=9)

        f.tight_layout(pad=1.08, h_pad=None, w_pad=None)
        f.savefig(os.path.join(dirp, filename), facecolor='w', edgecolor='w',
                  bbox_extra_artists=(plt.gci()), bbox_inches='tight', dpi=300)
        # PDF
        for tick in ax.xaxis.get_major_ticks():
            tick.label.set_fontsize(12)
        for tick in ax.yaxis.get_major_ticks():
            tick.label.set_fontsize(12)
        ax.xaxis.label.set_size(14)
        ax.yaxis.label.set_size(14)
        pp = PdfPages(os.path.splitext(os.path.join(dirp, filename))[0] + '.pdf')
        pp.savefig(f, bbox_extra_artists=(plt.gci()), bbox_inches='tight')
        pp.close()

    def gen_html(self, directory, parameters, ccf, align=50, alpha=0.05):
        dir_name = os.path.basename(directory)
        # check_dir(directory)
        html_header = "Promoter Test: " + dir_name
        self.link_d = OrderedDict()
        self.link_d["RNA"] = "index.html"
        self.link_d["Sig promoters"] = "spromoters.html"
        self.link_d["All promoters"] = "promoters.html"
        self.link_d["Parameters"] = "parameters.html"

        #############################################################
        # Index main page
        #############################################################
        html = Html(name=html_header, links_dict=self.link_d,  # fig_dir=os.path.join(directory,"style"),
                    fig_rpath="../style", RGT_header=False, other_logo="TDF", homepage="../index.html")

        html.add_figure("plot_promoter.png", align="left", width="45%", more_images=["bar_promoter.png"])

        if self.showdbs:
            html.add_figure("plot_dbss.png", align="left", width="45%", more_images=["bar_dbss.png"])

        # Table of merged TBS on promoters
        if self.showdbs:
            header_list = [["", "", "Promoters", None, None, None, None, None, "DBSs", None, None, None, None, None],
                           ["#", "DBD",
                            "Target Promoter", None, "Non-target Promoter", None, "Statistics", None,
                            "Target Promoter", None, "Non-target Promoter", None, "Statistics", None],
                           [" ", " ",
                            "with DBS", "without DBS", "with DBS", "without DBS", "OR", "<i>p</i>-value",
                            "No. DBSs", "Other DBSs", "No. DBSs", "Other DBSs", "OR", "<i>p</i>-value"]]
            header_titles = [["", "", "Statistics on promoter level", None, None, None, None, None,
                              "Statistics on DBS level", None, None, None, None, None],
                             ["Rank of the talbe",
                              "DNA Binding Domain which is the functional region on RNA.",
                              "Promoters of the differential expression genes.", None,
                              "Promoters of the non-differential expression genes.", None,
                              "Statistics based on promoters", None,
                              "Promoters of the differential expression genes.", None,
                              "Promoters of the non-differential expression genes.", None,
                              "Statistics based on DBSs", None],
                             ["",
                              "",
                              "Number of target promoters which contain DBSs (DNA Binding Sites).",
                              "Number of target promoters which don't contain DBSs (DNA Binding Sites).",
                              "Number of non-target promoters which contain DBSs (DNA Binding Sites).",
                              "Number of non-target promoters which don't contain DBSs (DNA Binding Sites).",
                              "Odds Ratio", "P-value",
                              "Number of DBSs found in the target promoters.",
                              "Number of DBSs not found in the target promoters.",
                              "Number of DBSs found in the non-target promoters.",
                              "Number of DBSs not found in the non-target promoters.",
                              "Odds Ratio", "P-value"]
                             ]
            border_list = [" style=\"border-right:1pt solid gray\"",
                           " style=\"border-right:1pt solid gray\"", "",
                           " style=\"border-right:1pt solid gray\"", "",
                           " style=\"border-right:1pt solid gray\"", "",
                           " style=\"border-right:2pt solid gray\"", "",
                           " style=\"border-right:1pt solid gray\"", "",
                           " style=\"border-right:1pt solid gray\"", "",
                           " style=\"border-right:1pt solid gray\""]
        else:
            header_list = [["#", "DBD", "Target Promoter", None, "Non-target Promoter", None, "Statistics", None],
                           [" ", " ", "with DBS", "without DBS", "with DBS", "without DBS", "OR", "<i>p</i>"]]
            header_titles = [["Rank of the talbe",
                              "DNA Binding Domain which is the functional region on RNA.",
                              "Promoters of the differential expression genes.", None,
                              "Promoters of the non-differential expression genes.", None,
                              "Statistics based on promoters", None],
                             ["",
                              "",
                              "Number of target promoters which contain DBSs (DNA Binding Sites).",
                              "Number of target promoters which don't contain DBSs (DNA Binding Sites).",
                              "Number of non-target promoters which contain DBSs (DNA Binding Sites).",
                              "Number of non-target promoters which don't contain DBSs (DNA Binding Sites).",
                              "Odds Ratio", "P-value"]
                             ]
            border_list = ["style=\"border-right:1pt solid gray\"",
                           "style=\"border-right:1pt solid gray\"", "",
                           "style=\"border-right:1pt solid gray\"", "",
                           "style=\"border-right:1pt solid gray\"", "",
                           "style=\"border-right:1pt solid gray\""]

        type_list = 'ssssssssssssssssssss'
        col_size_list = [20] * 20
        data_table = []
        rank = 0
        self.topDBD = ["-", 1]

        for rbs in self.frequency["promoters"]["de"].keys():
            if self.frequency["promoters"]["de"][rbs][0] < ccf: continue
            rank += 1
            if self.pvalue[rbs] < alpha:
                p_promoter = "<font color=\"red\">" + value2str(self.pvalue[rbs]) + "</font>"
            else:
                p_promoter = value2str(self.pvalue[rbs])

            if self.showdbs:
                if self.hpvalue[rbs] < alpha:
                    p_hit = "<font color=\"red\">" + value2str(self.hpvalue[rbs]) + "</font>"
                else:
                    p_hit = value2str(self.hpvalue[rbs])

            try:
                if self.pvalue[rbs] < self.topDBD[1]:
                    self.topDBD = [rbs.str_rna(pa=False), self.pvalue[rbs]]
            except:
                self.topDBD = [rbs.str_rna(pa=False), self.pvalue[rbs]]

            new_row = [str(rank),
                       rbs.str_rna(pa=False),
                       '<a href="' + "dbds_promoters.html#" + rbs.str_rna() +
                       '" style="text-align:left">' +
                       value2str(self.frequency["promoters"]["de"][rbs][0]) + '</a>',
                       value2str(self.frequency["promoters"]["de"][rbs][1]),
                       value2str(self.frequency["promoters"]["nde"][rbs][0]),
                       value2str(self.frequency["promoters"]["nde"][rbs][1]),
                       value2str(self.oddsratio[rbs]),
                       p_promoter]
            if self.showdbs:
                new_row += [value2str(self.frequency["hits"]["de"][rbs][0]),
                            value2str(self.frequency["hits"]["de"][rbs][1]),
                            value2str(self.frequency["hits"]["nde"][rbs][0]),
                            value2str(self.frequency["hits"]["nde"][rbs][1]),
                            value2str(self.hoddsratio[rbs]),
                            p_hit]

            data_table.append(new_row)
        # data_table = sorted(data_table, key=lambda x: x[-1])
        # data_table = sorted(data_table, key=lambda x: float(x[-1]))
        html.add_zebra_table(header_list, col_size_list, type_list, data_table, align=align, cell_align="left",
                             header_titles=header_titles, border_list=border_list, sortable=True)
        html.add_heading("Notes")
        html.add_list(["DBD stands for functional DNA Binding Domain on RNA.",
                       "RBS stands for RNA Binding Site on RNA.",
                       "DBS stands for DNA Binding Site on DNA."])

        #### Table for motif
        if self.motif:
            data_table = []
            header_list = ["DBD", "Motif"]
            for dbd, f in self.motif.iteritems():
                svg = '<object type="image/svg+xml" data="' + f + '">Your browser does not support SVG</object>'
                new_row = [dbd, svg]
                data_table.append(new_row)
            html.add_zebra_table(header_list, col_size_list, type_list, data_table, align=align, cell_align="left")

        ####
        html.add_fixed_rank_sortable()
        html.write(os.path.join(directory, "index.html"))

        #############################################################
        # RNA subpage: Profile of targeted promoters for each merged DNA Binding Domain
        #############################################################

        # dbds_promoters.html

        html = Html(name=html_header, links_dict=self.link_d,  # fig_dir=os.path.join(directory,"style"),
                    fig_rpath="../style", RGT_header=False, other_logo="TDF", homepage="../index.html")

        if self.scores:
            if self.scoreh and self.de_gene:
                score_header = [self.de_gene.cond[0]]
            else:
                # if "(" in self.scores.values()[0]:
                #     # scs = self.scores[0].replace("(", "")
                #     # scs = scs.replace(")", "")
                #     # scs = scs.split(",")
                #     # score_header = ["Score"] * len(scs)
                #     score_header = ["Fold_change", "Filtered"]
                #
                # else:
                score_header = ["Fold Change Score"]

            header_list = ["#", "Promoter", "Gene", "DBSs counts", "DBS coverage"]
            header_list += score_header
            header_list += ["Sum of Ranks"]
            header_titles = ["", "Target promoters", "Gene symbol",
                             "Number of DNA Binding sites locating within the promoter",
                             "The proportion of promoter covered by binding sites"]
            header_titles += ["Scores loaded by their absolute values from gene list or BED input. If there is annotation error for the gene names, it shows zero instead."] * len(
                score_header)
            header_titles += ["Sum up the ranks from left-hand side columns"]

        else:
            header_list = ["#", "Promoter", "Gene", "DBSs Count", "DBS coverage", "Sum of Ranks"]

            header_titles = ["", "Target promoters", "Gene symbol",
                             "Number of DNA Binding sites locating within the promoter",
                             "The proportion of promoter covered by binding sites",
                             "Sum up the ranks from left-hand side columns"]

        for rbsm in self.frequency["promoters"]["de"].keys():
            html.add_heading("DNA Binding Domain: " + rbsm.str_rna(), idtag=rbsm.str_rna())
            data_table = []
            # Calculate the ranking
            # rank_array
            rank_count = len(self.txp_de.merged_dict[rbsm]) - rank_array(
                [len(self.promoter["de"]["rd"][p.toString()]) for p in self.txp_de.merged_dict[rbsm]])
            rank_coverage = len(self.txp_de.merged_dict[rbsm]) - rank_array(
                [self.promoter["de"]["dbs_coverage"][p.toString()] for p in self.txp_de.merged_dict[rbsm]])

            if self.scores:
                # multiple_scores = False
                # if isinstance(self.scores[0], str):
                #     if "(" in self.scores[0]:
                #         def ranking(scores):
                #             rank_score = len(self.txp_de.merged_dict[rbsm]) - rank_array(scores)
                #             return rank_score
                #
                #         multiple_scores = True
                #         scs = []
                #         for i, promoter in enumerate(self.txp_de.merged_dict[rbsm]):
                #             s = self.de_gene.values[promoter.name.upper()].replace("(", "")
                #             s = s.replace(")", "")
                #             s = s.split(",")
                #             scs.append([float(ss) for ss in s])
                #         ar = numpy.array(scs)
                #         # ar.transpose()
                #         # print(ar)
                #         score_ar = ar.tolist()
                #         rank_score = numpy.apply_along_axis(ranking, axis=0, arr=ar)
                #         rank_score = rank_score.transpose()
                #         rank_sum = numpy.sum(rank_score, axis=0).tolist()
                #         rank_sum = [x + y + z for x, y, z in zip(rank_count, rank_coverage, rank_sum)]
                #         rank_score = rank_score.tolist()
                #
                #     else:
                #         # try:
                #         new_scores = []
                #         for i, promoter in enumerate(self.txp_de.merged_dict[rbsm]):
                #             try:
                #                 try:
                #                     try:
                #                         s = self.de_gene.values[promoter.name]
                #                     except:
                #                         s = self.de_gene.values[self.ensembl2symbol[promoter.name].upper()]
                #                 except:
                #                     new_scores.append(abs(float(s)))
                #             except:
                #                 new_scores.append(abs(self.score_dict[promoter.name]))
                #             if s == "Inf" or s == "inf":
                #                 new_scores.append(float("inf"))
                #             elif s == "-Inf" or s == "-inf":
                #                 new_scores.append(float("inf"))
                #             else:
                #                 try: new_scores.append(abs(float(s)))
                #                 except: new_scores.append(0)
                #
                #         scores = new_scores
                #         rank_score = len(self.txp_de.merged_dict[rbsm]) - rank_array(scores)
                #         rank_sum = [x + y + z for x, y, z in zip(rank_count, rank_coverage, rank_score)]
                # else:
                #     scores = [float(self.de_gene.values[p.name.upper()]) for p in self.txp_de.merged_dict[rbsm]]
                #     rank_score = len(self.txp_de.merged_dict[rbsm]) - rank_array(scores)
                #     rank_sum = [x + y + z for x, y, z in zip(rank_count, rank_coverage, rank_score)]
                rank_score = len(self.txp_de.merged_dict[rbsm]) - rank_array([self.scores[p.toString()] for p in self.txp_de.merged_dict[rbsm] ])
                rank_sum = [x + y + z for x, y, z in zip(rank_count, rank_coverage, rank_score)]

            else:
                rank_sum = [x + y for x, y in zip(rank_count, rank_coverage)]

            for i, promoter in enumerate(self.txp_de.merged_dict[rbsm]):
                # Add information
                region_link = region_link_internet(self.organism, promoter)
                try:
                    newline = [str(i + 1), region_link,
                               split_gene_name(gene_name=self.ensembl2symbol[promoter.name], org=self.organism),
                               str(len(self.promoter["de"]["rd"][promoter.toString()])),
                               value2str(self.promoter["de"]["dbs_coverage"][promoter.toString()])
                               ]
                except:
                    newline = [str(i + 1), region_link,
                               split_gene_name(gene_name=promoter.name, org=self.organism),
                               str(len(self.promoter["de"]["rd"][promoter.toString()])),
                               value2str(self.promoter["de"]["dbs_coverage"][promoter.toString()])
                               ]
                if self.scores:
                    # if multiple_scores:
                    #     for j in range(len(score_ar[0])):
                    #
                    #         newline += [value2str(score_ar[i][j])]
                    # else:
                    newline += [value2str(abs(self.scores[promoter.toString()]))]

                newline += ["<i>" + str(int(rank_sum[i])) + "</i>"]
                # print(newline)
                data_table.append(newline)
            data_table = sorted(data_table, key=lambda x: x[-1])
            html.add_zebra_table(header_list, col_size_list, type_list, data_table, align=align, cell_align="left",
                                 header_titles=header_titles, sortable=True, border_list=None)
        html.add_fixed_rank_sortable()
        html.write(os.path.join(directory, "dbds_promoters.html"))

        ################################################################
        ############# Parameters
        html = Html(name=html_header, links_dict=self.link_d,  # fig_dir=os.path.join(directory,"style"),
                    fig_rpath="../style", RGT_header=False, other_logo="TDF", homepage="../index.html")

        html.add_heading("Parameters")
        header_list = ["Description", "Arguments", "Value"]

        if parameters.de:
            de = os.path.basename(parameters.de)
            bed = "False"
            bg = "False"
        else:
            de = "False"
            bed = os.path.basename(parameters.bed)
            bg = os.path.basename(parameters.bg)

        data_table = [["RNA sequence name", "-rn", parameters.rn],
                      ["Input RNA sequence file", "-r", os.path.basename(parameters.r)],
                      ["Input file for defferentially expression gene list", "-de", de],
                      ["Input BED file as promoters", "-bed", bed],
                      ["Input BED file as backgrounds", "-bg", bg],
                      ["Output directory", "-o", "/".join(parameters.o.partition("/")[-3:])],
                      ["Organism", "-organism", parameters.organism],
                      ["filter_havana", "-filter_havana", parameters.filter_havana],
                      ["protein_coding", "-protein_coding", parameters.protein_coding],
                      ["known_only", "-known_only", parameters.known_only],
                      ["Promoter length", "-pl", str(parameters.pl)],
                      ["Alpha level for rejection p value", "-a", str(parameters.a)],
                      ["Cut off value for filtering out the low counts of DBSs", "-ccf", str(parameters.ccf)],
                      ["Remove temporary files", "-rt", str(parameters.rt)],
                      ["Input file for RNA accecibility", "-ac", str(parameters.ac)],
                      ["Cut off value for RNA accecibility", "-accf", str(parameters.accf)],
                      ["Output the BED files for DNA binding sites.", "-obed", str(parameters.obed)],
                      ["Show parallel and antiparallel bindings in the plot separately.", "-showpa",
                       str(parameters.showpa)],
                      ["Minimum length", "-l", str(self.triplexator_p[0])],
                      ["Maximum error rate", "-e", str(self.triplexator_p[1])],
                      ["Tolerated number of consecutive errors", "-c", str(self.triplexator_p[2])],
                      ["Filtering repeats", "-fr", str(self.triplexator_p[3])],
                      ["Filtering mode", "-fm", str(self.triplexator_p[4])],
                      ["Output format", "-of", str(self.triplexator_p[5])],
                      ["Merge features", "-mf", str(self.triplexator_p[6])]]
        html.add_zebra_table(header_list, col_size_list, type_list, data_table, align=align, cell_align="left",
                             auto_width=True)
        html.add_free_content(['<a href="summary.txt" style="margin-left:100">See details</a>'])
        html.write(os.path.join(directory, "parameters.html"))

    def gen_html_genes(self, directory, align=50, alpha=0.05, nonDE=False):
        dir_name = os.path.basename(directory)
        html_header = "Promoter Test: " + dir_name
        self.ranktable = {}
        self.dbstable = {}
        type_list = 'sssssssssssssss'
        col_size_list = [10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10]

        #############################################################
        # Promoter centered
        #############################################################

        ##############################################################################################
        # promoters.html
        html = Html(name=html_header, links_dict=self.link_d,  # fig_dir=os.path.join(directory,"style"),
                    fig_rpath="../style", RGT_header=False, other_logo="TDF", homepage="../index.html")

        if self.scores:
            if self.scoreh and self.de_gene:
                score_header = [self.de_gene.cond[0]]
            else:
                # if "(" in self.scores.values()[0]:
                #     score_header = ["Fold_change", "Filtered"]
                # else:
                score_header = ["Fold Change Score"]
            header_listp = ["#", "Promoter", "Gene", "DBSs Count", "DBS coverage"]
            header_listp += score_header
            header_listp += ["Sum of Ranks"]

            header_titlesp = ["", "Target promoters", "Gene symbol",
                              "Number of DNA Binding sites locating within the promoter",
                              "The proportion of promoter covered by binding sites"]
            header_titlesp += ["Scores loaded by their absolute values from gene list or BED input. If there is annotation error for the gene names, it shows zero instead."] * len(
                score_header)
            header_titlesp += ["Sum up the ranks from left-hand side columns"]

        else:
            header_listp = ["#", "Promoter", "Gene", "DBSs Count", "DBS coverage", "Sum of Ranks"]

            header_titlesp = ["", "Target promoters", "Gene symbol",
                              "Number of DNA Binding sites locating within the promoter",
                              "The proportion of promoter covered by binding sites",
                              "Sum up the ranks from left-hand side columns"]

        html.add_heading("Target promoters")
        data_table = []

        if not self.de_regions.sorted: self.de_regions.sort()
        # Iterate by each gene promoter

        # Calculate the ranking
        rank_count = len(self.de_regions) - rank_array(
            [self.promoter["de"]["dbs"][p.toString()] for p in self.de_regions])
        rank_coverage = len(self.de_regions) - rank_array(
            [self.promoter["de"]["dbs_coverage"][p.toString()] for p in self.de_regions])

        if self.scores:
            # multiple_scores = False
            # if isinstance(self.scores[0], str):
            #     new_scores = []
            #     if "(" in self.scores[0]:
            #         def ranking(scores):
            #             rank_score = len(self.de_regions) - rank_array(scores)
            #             return rank_score
            #
            #         multiple_scores = True
            #         # scs = []
            #         for s in self.scores:
            #             s = s.replace("(", "")
            #             s = s.replace(")", "")
            #             s = s.split(",")
            #             new_scores.append([float(ss) for ss in s])
            #         ar = numpy.array(new_scores)
            #         # ar.transpose()
            #         # print(ar)
            #         score_ar = ar.tolist()
            #         rank_score = numpy.apply_along_axis(ranking, axis=0, arr=ar)
            #         rank_score = rank_score.transpose()
            #         rank_sum = numpy.sum(rank_score, axis=0).tolist()
            #         rank_sum = [x + y + z for x, y, z in zip(rank_count, rank_coverage, rank_sum)]
            #         rank_score = rank_score.tolist()
            #         scores = new_scores
            #
            #     else:
            #
            #         for s in self.scores:
            #             if s == "Inf" or s == "inf":
            #                 new_scores.append(float("inf"))
            #             elif s == "-Inf" or s == "-inf":
            #                 new_scores.append(-float("inf"))
            #             else:
            #                 new_scores.append(abs(float(s)))
            #
            #         scores = new_scores
            #         rank_score = len(self.de_regions) - rank_array(scores)
            #         rank_sum = [x + y + z for x, y, z in zip(rank_count, rank_coverage, rank_score)]
            # else:
            #     rank_score = len(self.de_regions) - rank_array(scores)
            #     rank_sum = [x + y + z for x, y, z in zip(rank_count, rank_coverage, rank_score)]

            rank_score = len(self.de_regions) - rank_array([self.scores[p.toString()] for p in self.de_regions])
            rank_sum = [x + y + z for x, y, z in zip(rank_count, rank_coverage, rank_score)]

        else:
            rank_sum = [x + y for x, y in zip(rank_count, rank_coverage)]

        for i, promoter in enumerate(self.de_regions):

            if self.promoter["de"]["dbs"][promoter.toString()] == 0:
                dbssount = str(0)
            else:
                dbssount = '<a href="promoters_dbds.html#' + promoter.toString() + '" style="text-align:left">' + \
                           str(self.promoter["de"]["dbs"][promoter.toString()]) + '</a>'

            region_link = region_link_internet(self.organism, promoter)

            try:
                gn = self.ensembl2symbol[promoter.name]
                if not gn: gn = promoter.name
            except:
                gn = promoter.name

            self.ranktable[gn] = str(int(rank_sum[i]))
            self.dbstable[gn] = str(int(self.promoter["de"]["dbs"][promoter.toString()]))

            newline = [str(i + 1),
                       region_link,
                       split_gene_name(gene_name=gn, org=self.organism),
                       dbssount,
                       value2str(self.promoter["de"]["dbs_coverage"][promoter.toString()])
                       ]

            if self.scores:
                # if multiple_scores:
                #     for j in range(len(score_ar[0])):
                #         newline += [value2str(abs(score_ar[i][j]))]
                # else:
                newline += [value2str(abs(self.scores[promoter.toString()]))]

            newline += ["<i>" + str(int(rank_sum[i])) + "</i>"]
            # print(newline)
            data_table.append(newline)

        # print(data_table)
        data_table = natsort.natsorted(data_table, key=lambda x: x[-1])
        html.add_zebra_table(header_listp, col_size_list, type_list, data_table, align=align, cell_align="left",
                             header_titles=header_titlesp, border_list=None, sortable=True)
        html.add_heading("Notes")
        html.add_list(["DBS stands for DNA Binding Site on DNA.",
                       "DBS coverage is the proportion of the promoter where has potential to form triple helices with the given RNA."])
        html.add_fixed_rank_sortable()
        html.write(os.path.join(directory, "promoters.html"))

        ############################
        # Subpages for promoter centered page
        # promoters_dbds.html
        header_sub = ["#", "RBS", "DBS", "Strand", "Score", "Motif", "Orientation"]
        header_titles = ["", "RNA Binding Site", "DNA Binding Site", "Strand of DBS on DNA",
                         "Score of binding event", "Motif of binding by triple helix rule",
                         "Orientation of interaction between DNA and RNA. 'P'- Parallel; 'A'-Antiparallel"]
        header_list = header_sub
        html = Html(name=html_header, links_dict=self.link_d,  # fig_dir=os.path.join(directory,"style"),
                    fig_rpath="../style", RGT_header=False, other_logo="TDF", homepage="../index.html")

        for i, promoter in enumerate(self.de_regions):
            if self.promoter["de"]["dbs"][promoter.toString()] == 0:
                continue
            else:
                try:
                    gn = self.ensembl2symbol[promoter.name]
                except:
                    gn = promoter.name
                html.add_heading(split_gene_name(gene_name=gn, org=self.organism), idtag=promoter.toString())
                html.add_free_content(['<a href="http://genome.ucsc.edu/cgi-bin/hgTracks?db=' + self.organism +
                                       "&position=" + promoter.chrom + "%3A" + str(promoter.initial) + "-" + str(
                    promoter.final) +
                                       '" style="margin-left:50">' +
                                       promoter.toString(space=True) + '</a>'])
                data_table = []

                for j, rd in enumerate(self.promoter["de"]["rd"][promoter.toString()]):
                    rbs = rd.rna.str_rna(pa=False)
                    for rbsm in self.sig_DBD:
                        # rbsm = rbsm.partition(":")[2].split("-")
                        if rd.rna.overlap(rbsm):
                            rbs = "<font color=\"red\">" + rbs + "</font>"

                    data_table.append([str(j + 1), rbs, rd.dna.toString(space=True),
                                       rd.dna.orientation, rd.score, rd.motif, rd.orient])

                html.add_zebra_table(header_list, col_size_list, type_list, data_table, align=align, cell_align="left",
                                     header_titles=header_titles, sortable=True)
        html.add_fixed_rank_sortable()
        html.write(os.path.join(directory, "promoters_dbds.html"))

        ##############################################################################################
        # spromoters.html    for significant promoters
        spromoters = GenomicRegionSet("sig_promoters")
        self.promoter["de"]["sig_dbs"] = {}
        self.promoter["de"]["sig_dbs_coverage"] = {}
        for promoter in self.de_regions:
            # for rd in self.promoter["de"]["rd"][promoter.toString()]:
            #     if rd.rna
            sig_bindings = self.promoter["de"]["rd"][promoter.toString()].overlap_rbss(rbss=self.sig_DBD)
            dbs = sig_bindings.get_dbs()
            if len(dbs) > 0:
                spromoters.add(promoter)
                m_dbs = dbs.merge(w_return=True)
                self.promoter["de"]["sig_dbs"][promoter.toString()] = len(dbs)
                # self.promoter["de"]["merged_dbs"][promoter.toString()] = len(m_dbs)
                self.promoter["de"]["sig_dbs_coverage"][promoter.toString()] = float(m_dbs.total_coverage()) / len(promoter)

        html = Html(name=html_header, links_dict=self.link_d,  # fig_dir=os.path.join(directory,"style"),
                    fig_rpath="../style", RGT_header=False, other_logo="TDF", homepage="../index.html")

        # Select promoters in sig DBD
        # spromoters = self.promoter["de"]["sig_dbs"].keys()
        if len(spromoters) == 0:
            html.add_heading("There is no significant DBD.")
        else:
            html.add_heading("Target promoters bound by significant DBD")
            # for rbsm in self.sig_region_promoter:
            #     spromoters = spromoters + [p for p in self.txp_de.merged_dict[rbsm]]
            # spromoters = list(set(spromoters))
            data_table = []

            # Iterate by each gene promoter

            # Calculate the ranking
            rank_count = len(spromoters) - rank_array([self.promoter["de"]["sig_dbs"][p.toString()] for p in spromoters])
            rank_coverage = len(spromoters) - rank_array(
                [self.promoter["de"]["sig_dbs_coverage"][p.toString()] for p in spromoters])

            if self.scores:
                # sscores = []
                # de_genes_str = [g.name for g in self.de_gene.genes]
                # for p in spromoters:
                #     try: gene_sym = self.ensembl2symbol[p.name].upper()
                #     except: gene_sym = p.name.upper()
                #     try: sscores.append(self.de_gene.values[gene_sym.upper()])
                #     except: sscores.append(0)
                #
                # if isinstance(sscores[0], str):
                #
                #         new_scores = []
                #         for s in sscores:
                #             if s == "Inf" or s == "inf":
                #                 new_scores.append(float("inf"))
                #             elif s == "-Inf" or s == "-inf":
                #                 new_scores.append(-float("inf"))
                #             elif isinstance(s, str):
                #                 new_scores.append(s)
                #             else:
                #                 new_scores.append(abs(float(s)))
                #
                #         scores = new_scores
                #         rank_score = len(spromoters) - rank_array(scores)
                #         rank_sum = [x + y + z for x, y, z in zip(rank_count, rank_coverage, rank_score)]
                #
                # else:
                rank_score = len(spromoters) - rank_array([self.scores[p.toString()] for p in spromoters ] )
                rank_sum = [x + y + z for x, y, z in zip(rank_count, rank_coverage, rank_score)]

            else:
                rank_sum = [x + y for x, y in zip(rank_count, rank_coverage)]

            for i, promoter in enumerate(spromoters):
                try:
                    gn = self.ensembl2symbol[promoter.name]
                except:
                    gn = promoter.name
                dbssount = '<a href="promoters_dbds.html#' + promoter.toString() + \
                           '" style="text-align:left">' + \
                           str(self.promoter["de"]["sig_dbs"][promoter.toString()]) + '</a>'

                region_link = region_link_internet(self.organism, promoter)

                self.ranktable[gn] = str(int(rank_sum[i]))
                self.dbstable[gn] = str(self.promoter["de"]["sig_dbs"][promoter.toString()])

                newline = [str(i + 1),
                           region_link,
                           split_gene_name(gene_name=gn, org=self.organism),
                           dbssount,
                           value2str(self.promoter["de"]["sig_dbs_coverage"][promoter.toString()])
                           ]
                if self.scores:
                    # if multiple_scores:
                    #     # print(score_ar)
                    #     # print(len(score_ar))
                    #     for j in range(len(score_ar[0])):
                    #         # print(score_ar[j][i])
                    #         # print(score_ar[i][j])
                    #         newline += [value2str(score_ar[i][j])]
                    # else:
                    newline += [value2str(self.scores[promoter.toString()])]

                newline += [str(int(rank_sum[i]))]
                # print(newline)
                data_table.append(newline)

            data_table = natsort.natsorted(data_table, key=lambda x: x[-1])
            html.add_zebra_table(header_listp, col_size_list, type_list, data_table, align=align, cell_align="left",
                                 header_titles=header_titlesp, border_list=None, sortable=True)
            html.add_heading("Notes")
            html.add_list(["DBS stands for DNA Binding Site on DNA.",
                           "DBS coverage is the proportion of the promoter where has potential to form triple helices with the given RNA."])
            html.add_fixed_rank_sortable()
        html.write(os.path.join(directory, "spromoters.html"))



    def save_table(self, path, table, filename):
        """Save the summary rank into the table for heatmap"""

        table_path = os.path.join(path, filename)
        # self.ranktable = {}
        rank_table = []

        if os.path.isfile(table_path):
            # Table exists
            f = open(table_path)
            for i, line in enumerate(f):
                line = line.strip().split()
                if i == 0:
                    # if not line:
                    #     break
                    #     exist = False
                    if self.rna_name in line:
                        # lncRNA exists
                        exist = True
                        ind_rna = line.index(self.rna_name)
                        header = line
                    else:
                        exist = False
                        line.append(self.rna_name)
                        header = line

                else:
                    if exist and ind_rna:
                        # try:
                        line[ind_rna + 1] = table[line[0]]
                        rank_table.append(line)
                        # except:
                        #     rank_table.append(line)
                    else:
                        # try:
                        line.append(table[line[0]])
                        rank_table.append(line)
                        # except:
                        #     rank_table.append(line)
            f.close()

        else:
            # Table not exists
            header = ["gene", self.rna_name]
            for k, v in table.iteritems():
                rank_table.append([k, v])

        # Write into file
        g = open(table_path, "w")

        print("\t".join(header), file=g)
        for l in rank_table:
            # print(l)
            print("\t".join(l), file=g)
        g.close()

    def heatmap(self, table, temp):
        """Generate heatmap for comparisons among genes and lncRNAs"""

        # Generate random features and distance matrix.
        genes = []
        data = []
        if not os.path.isfile(os.path.join(temp, table)):
            print("*** No rank table is found.")
            return

        f = open(os.path.join(temp, table))

        for i, line in enumerate(f):
            line = line.strip().split()
            if i == 0:
                rnas = line[1:]
            else:
                genes.append(line[0])
                data.append([int(v) for v in line[1:]])

        f.close()
        data = numpy.array(data)
        # print(type(data))
        # print(data.shape)

        # Compute and plot first dendrogram.
        fig = plt.figure(figsize=(max(20, 4 + int(data.shape[1] * 1.6)),
                                  max(80, int(data.shape[0] * 0.2))))
        # fig.suptitle("Heatmap of summary ranks", fontsize=20, y=0.95)

        ax1 = fig.add_axes([0.09, 0.05, 0.2, 0.9])
        Y = sch.linkage(data, method='single')
        Z1 = sch.dendrogram(Y, orientation='right')
        ax1.set_xticks([])
        ax1.set_yticks([])

        # Compute and plot second dendrogram.
        ax2 = fig.add_axes([0.3, 1.1, 0.55, 0.04])
        # tdata = numpy.transpose(data)
        Y = sch.linkage(data.T, method='single')
        Z2 = sch.dendrogram(Y)
        ax2.set_xticks([])
        ax2.set_yticks([])

        # Plot distance matrix.
        axmatrix = fig.add_axes([0.3, 0.05, 0.55, 0.9])
        # axmatrix = fig.add_axes()
        idx1 = Z1['leaves']
        idx2 = Z2['leaves']
        data = data[idx1, :]
        data = data[:, idx2]
        im = axmatrix.matshow(data, aspect='auto', origin='lower', cmap=plt.cm.YlGnBu)

        axmatrix.set_xticks(range(data.shape[1]))
        axmatrix.set_xticklabels([rnas[i] for i in idx2], minor=False, ha="left")
        axmatrix.xaxis.set_label_position('top')
        axmatrix.xaxis.tick_top()
        plt.xticks(rotation=70, fontsize=12)

        axmatrix.set_yticks(range(data.shape[0]))
        axmatrix.set_yticklabels([genes[i] for i in idx1], minor=False)
        axmatrix.yaxis.set_label_position('right')
        axmatrix.yaxis.tick_right()
        plt.yticks(rotation=0, fontsize=12)

        # Plot colorbar.
        axcolor = fig.add_axes([0.1, 0.02, 0.8, 0.01])
        plt.colorbar(im, cax=axcolor, orientation='horizontal')
        axcolor.set_xlabel('summary rank')
        fig.savefig(os.path.join(temp, 'lncRNA_target_dendrogram.png'))
        fig.savefig(os.path.join(temp, 'lncRNA_target_dendrogram.pdf'), format="pdf")





# Import
import os
import sys
from glob import glob
from Bio import motifs
from os.path import basename

# Input
inFolder = sys.argv[1]
outFileName = sys.argv[2]

# Parameters
fprList = [0.005, 0.001, 0.0005, 0.0001, 0.00005, 0.00001]
pseudocounts = 0.1
background = {'A':0.25,'C':0.25,'G':0.25,'T':0.25}
precision = 10000

# Creating output file
outFile = open(outFileName,"w")
outFile.write("\t".join(["MOTIF"]+[str(e) for e in fprList])+"\n")

# Iterating on all PWMs
for pwmFileName in glob(inFolder+"*.pwm"):

  # Creating PSSM
  name = ".".join(basename(pwmFileName).split(".")[:-1])
  input_file = open(pwmFileName,"r")
  pfm = motifs.read(input_file, "pfm")
  pwm = pfm.counts.normalize(pseudocounts)
  input_file.close()
  pssm = pwm.log_odds(background)
  pssm_list = [pssm[e] for e in ["A","C","G","T"]]
  distribution = pssm.distribution(background=background, precision=precision)

  # Evaluating thresholds
  resVec = [name]
  for fpr in fprList:
    resVec.append(str(distribution.threshold_fpr(fpr)))
    
  # Writing results
  outFile.write("\t".join(resVec)+"\n")





###################################################################################################
# Libraries
###################################################################################################

# Python
import os
import numpy as np
from collections import Counter

# Internal
from ..Util import GenomeData
from signalProcessing import GenomicSignal
from rgt.GenomicRegionSet import GenomicRegionSet
from hmm import HMM
from biasTable import BiasTable

"""
Train a hidden Markov model (HMM) based on the annotation data

Authors: Eduardo G. Gusmao, Zhijian Li
"""


class TrainHMM:
    """
    Contains methods used to train a hidden Markov model
    """

    def __init__(self, bam_file, annotate_file, print_bed_file,
                 output_locaiton, output_fname,
                 print_raw_signal, print_bc_signal, print_norm_signal, print_slope_signal,
                 atac_initial_clip, atac_downstream_ext, atac_upstream_ext,
                 atac_forward_shift, atac_reverse_shift,
                 estimate_bias_correction, estimate_bias_type, bias_table,
                 original_regions, organism, k_nb):
        self.bam_file = bam_file
        self.annotate_fname = annotate_file
        self.print_bed_file = print_bed_file
        self.output_locaiton = output_locaiton
        self.output_fname = output_fname
        self.print_raw_signal = print_raw_signal
        self.print_bc_signal = print_bc_signal
        self.print_norm_signal = print_norm_signal
        self.print_slope_signal = print_slope_signal
        self.atac_initial_clip = atac_initial_clip
        self.atac_downstream_ext = atac_downstream_ext
        self.atac_upstream_ext = atac_upstream_ext
        self.atac_forward_shift = atac_forward_shift
        self.atac_reverse_shift = atac_reverse_shift
        self.estimate_bias_correction = estimate_bias_correction
        self.estimate_bias_type = estimate_bias_type
        self.bias_table = bias_table
        self.original_regions = original_regions
        self.organism = organism
        self.k_nb = k_nb
        self.chrom = "chr1"
        self.start = 211428000
        self.end = 211438000

    def read_states_signals(self):
        # Read states from the annotation file
        states = ""
        with open(self.annotate_fname) as annotate_file:
            for line in annotate_file:
                if len(line) < 2 or "#" in line or "=" in line:
                    continue
                ll = line.strip().split(" ")
                for state in ll[1:-1]:
                    states += state

        # If need to estimate bias table
        bias_table = BiasTable(output_loc=self.output_locaiton)
        genome_data = GenomeData(self.organism)
        table = None
        if self.estimate_bias_correction:
            regions = GenomicRegionSet("Bias Regions")
            if self.original_regions.split(".")[-1] == "bed":
                regions.read_bed(self.original_regions)
            if self.original_regions.split(".")[-1] == "fa":
                regions.read_sequence(self.original_regions)

            if self.estimate_bias_type == "FRE":
                table = bias_table.estimate_table(regions=regions, dnase_file_name=self.bam_file,
                                                  genome_file_name=genome_data.get_genome(),
                                                  k_nb=self.k_nb,
                                                  forward_shift=self.atac_forward_shift,
                                                  reverse_shift=self.atac_reverse_shift)
            elif self.estimate_bias_type == "PWM":
                table = bias_table.estimate_table_pwm(regions=regions, dnase_file_name=self.bam_file,
                                                      genome_file_name=genome_data.get_genome(),
                                                      k_nb=self.k_nb,
                                                      forward_shift=self.atac_forward_shift,
                                                      reverse_shift=self.atac_reverse_shift)

            bias_fname = os.path.join(self.output_locaiton, "Bias", "{}_{}".format(self.k_nb, self.atac_forward_shift))
            bias_table.write_tables(bias_fname, table)

        # If the bias table is provided
        if self.bias_table:
            bias_table_list = self.bias_table.split(",")
            table = bias_table.load_table(table_file_name_F=bias_table_list[0],
                                          table_file_name_R=bias_table_list[1])

        # Get the normalization and slope signal from the raw bam file
        raw_signal = GenomicSignal(self.bam_file)
        raw_signal.load_sg_coefs(slope_window_size=9)
        norm_signal, slope_signal = raw_signal.get_signal(ref=self.chrom, start=self.start, end=self.end,
                                                          downstream_ext=self.atac_downstream_ext,
                                                          upstream_ext=self.atac_upstream_ext,
                                                          forward_shift=self.atac_forward_shift,
                                                          reverse_shift=self.atac_reverse_shift,
                                                          initial_clip=self.atac_initial_clip,
                                                          bias_table=table,
                                                          genome_file_name=genome_data.get_genome(),
                                                          print_raw_signal=self.print_raw_signal,
                                                          print_bc_signal=self.print_bc_signal,
                                                          print_norm_signal=self.print_norm_signal,
                                                          print_slope_signal=self.print_slope_signal)
        if self.print_bed_file:
            self.output_bed_file(states)

        return states, norm_signal, slope_signal

    def train(self):
        # Estimate the HMM parameters using the maximum likelihood method.
        states, norm_signal, slope_signal = self.read_states_signals()
        hmm_model = HMM()

        hmm_model.dim = 2
        # States number
        state_list = [int(state) for state in list(states)]
        hmm_model.states = len(np.unique(np.array(state_list)))

        # Initial state probabilities vector
        init_state = state_list[0]
        hmm_model.pi = [0.0] * hmm_model.states
        hmm_model.pi[init_state] = 1.0

        # Transition
        trans_matrix = np.zeros((hmm_model.states, hmm_model.states))
        for (x, y), c in Counter(zip(state_list, state_list[1:])).iteritems():
            trans_matrix[x, y] = c

        for i in range(hmm_model.states):
            trans_list = list()
            for j in range(hmm_model.states):
                trans_list.append(trans_matrix[i][j])
            trans_sum = sum(trans_list) * 1.0
            prob_list = [e / trans_sum for e in trans_list]

            # make sure that the sum of this line converge to 1
            total_prob = sum(prob_list)
            if total_prob != 1.0:
                prob_list[0] += (1.0 - total_prob)

            hmm_model.A.append(prob_list)

        # Emission
        for i in range(hmm_model.states):
            norm = list()
            slope = list()
            for j in range(len(state_list)):
                if state_list[j] == i:
                    norm.append(norm_signal[j])
                    slope.append(slope_signal[j])
            # Compute the mean of norm and slope signal
            means_list = list()
            means_list.append(np.mean(norm))
            means_list.append(np.mean(slope))
            hmm_model.means.append(means_list)

            # Compute covariance matrix of norm and slope signal
            covs_list = list()
            covs_matrix = np.cov(norm, slope)
            for j in range(hmm_model.dim):
                for k in range(hmm_model.dim):
                    covs_list.append(covs_matrix[j][k] + 0.000001) # covariance must be symmetric, positive-definite
            hmm_model.covs.append(covs_list)

        if self.estimate_bias_correction:
            model_fname = os.path.join(self.output_locaiton, "Model", "{}_{}".format(self.k_nb, self.atac_forward_shift))
        else:
            model_fname = os.path.join(self.output_locaiton, "Model", self.output_fname)
        hmm_model.save_hmm(model_fname)

    def output_bed_file(self, states):
        bed_fname = os.path.join(self.output_locaiton, "Model", self.output_fname)
        bed_fname += ".bed"

        state_dict = dict([(0, "BACK"), (1, "UPD"), (2, "TOPD"), (3, "DOWND"), (4, "FP")])
        color_dict = dict([(0, "50,50,50"), (1, "10,80,0"), (2, "20,40,150"), (3, "150,20,40"), (4, "198,150,0")])

        state_list = [int(state) for state in list(states)]
        current_state = state_list[0]
        start_postion = self.start
        is_print = False
        with open(bed_fname, "w") as bed_file:
            for i in range(len(state_list)):
                if state_list[i] != current_state:
                    end_position = self.start + i
                    is_print = True
                elif i == len(state_list) - 1:
                    end_position = self.end
                    is_print = True

                if is_print:
                    bed_file.write(self.chrom + " " + str(start_postion) + " " + str(end_position) + " "
                                   + state_dict[current_state] + " " + str(1000) + " . "
                                   + str(start_postion) + " " + str(end_position) + " "
                                   + color_dict[current_state] + "\n")
                    start_postion = end_position
                    current_state = state_list[i]
                    is_print = False
